"""
Discovery plugin to create PanoptesResources from JSON config files.
"""
import json

from yahoo_panoptes.discovery.panoptes_discovery_plugin import PanoptesDiscoveryPlugin, PanoptesDiscoveryPluginError
from yahoo_panoptes.framework.plugins.context import PanoptesPluginContext
from yahoo_panoptes.framework.resources import PanoptesResource, PanoptesResourceSet


class PluginDiscoveryJSONFile(PanoptesDiscoveryPlugin):
    """
    Standalone discovery plugin to populate PanoptesResources from a JSON file.
    """
    def run(self, context):
        """
        The main entry point to the plugin

        Args:
            context (PanoptesPluginContext): The Plugin Context passed by the Plugin Agent

        Returns:
            PanoptesResourceSet: A non-empty resource set

        Raises:
            PanoptesDiscoveryPluginError: This exception is raised if any part of the lookup process has errors
        """

        assert context and isinstance(context, PanoptesPluginContext), 'context must be a PanoptesPluginContext'

        conf = context.config
        logger = context.logger
        config_file = None

        try:
            config_file = conf['main']['config_file']
            with open(config_file) as f:
                resource_specs = json.load(f)
        except Exception as e:
            raise PanoptesDiscoveryPluginError(
                'Error while attempting to parse JSON from file {}: {}'.format(config_file, repr(e))
            )

        resources = PanoptesResourceSet()
        num_successes = 0
        num_failures = 0

        for resource_spec in resource_specs:
            try:
                resource = PanoptesResource.resource_from_dict(resource_spec)
                resources.add(resource)
                num_successes += 1
                logger.debug('Added resource {} from JSON file {}'.format(resource, config_file))
            except Exception as e:
                logger.debug('Error while attempting to create a PanoptesResource from file {}: {}'.format(
                    config_file, repr(e)))
                num_failures += 1
                continue

        if num_successes > 0:
            logger.info('Tried to read {} resources from {}, {} failed'.format(num_successes + num_failures,
                                                                               config_file,
                                                                               num_failures))
        else:
            logger.error('Error while attempting to create PanoptesResources from {}.'.format(config_file))
            raise PanoptesDiscoveryPluginError(
                    'Error during lookup for PanoptesResource from file {}.'.format(config_file))

        return resources



import numbers
import time

from cached_property import threaded_cached_property

from yahoo_panoptes.polling.polling_plugin import PanoptesPollingPlugin
from yahoo_panoptes.framework.metrics import PanoptesMetric, PanoptesMetricType, PanoptesMetricsGroup, \
    PanoptesMetricsGroupSet, PanoptesMetricDimension
from yahoo_panoptes.framework.plugins.base_snmp_plugin import PanoptesSNMPBasePlugin
from yahoo_panoptes.framework.utilities.snmp.mibs.dot3StatsTable import *
from yahoo_panoptes.framework.utilities.snmp.mibs.ifTable import *
from yahoo_panoptes.framework.utilities.snmp.mibs.ifXTable import *
from yahoo_panoptes.framework.validators import PanoptesValidators
from yahoo_panoptes.plugins.polling.utilities.polling_status import PanoptesPollingStatus


class _INTERFACE_STATES(object):
    UP, DOWN, UNKNOWN = range(3)


_MAX_REPETITIONS = 25
_MISSING_METRIC_VALUE = -1
_DEFAULT_DIMENSION_VALUE = '<not set>'

_METRIC_TYPE_MAP = {
    'mtu': PanoptesMetricType.GAUGE,
    'admin_state': PanoptesMetricType.GAUGE,
    'oper_state': PanoptesMetricType.GAUGE,
    'oper_admin_state_mismatch': PanoptesMetricType.GAUGE,
    'configured_speed': PanoptesMetricType.GAUGE,
    'type': PanoptesMetricType.COUNTER,
    'errors_in': PanoptesMetricType.COUNTER,
    'errors_out': PanoptesMetricType.COUNTER,
    'discards_in': PanoptesMetricType.COUNTER,
    'discards_out': PanoptesMetricType.COUNTER,
    'bits_in': PanoptesMetricType.COUNTER,
    'bits_out': PanoptesMetricType.COUNTER,
    'unicast_packets_in': PanoptesMetricType.COUNTER,
    'unicast_packets_out': PanoptesMetricType.COUNTER,
    'multicast_packets_in': PanoptesMetricType.COUNTER,
    'multicast_packets_out': PanoptesMetricType.COUNTER,
    'broadcast_packets_in': PanoptesMetricType.COUNTER,
    'broadcast_packets_out': PanoptesMetricType.COUNTER,
    'total_packets_in': PanoptesMetricType.COUNTER,
    'total_packets_out': PanoptesMetricType.COUNTER,
    'errors_frame': PanoptesMetricType.COUNTER,
    'errors_crc': PanoptesMetricType.COUNTER,
    'errors_giants': PanoptesMetricType.COUNTER
}


class PluginPollingDeviceInterfaceMetrics(PanoptesSNMPBasePlugin, PanoptesPollingPlugin):
    def __init__(self):
        super(PluginPollingDeviceInterfaceMetrics, self).__init__()
        self._device_interface_metrics = PanoptesMetricsGroupSet()
        self._polling_status = None
        self._interface_metrics_group = None

        self._dot3stats_map = None
        self._if_table_stats_map = None
        self._ifx_table_stats_map = None

        self._DIMENSION_MAP = {
            'alias': self.get_alias,
            'media_type': self.get_media_type,
            'description': self.get_description,
            'configured_speed': self.get_configured_speed,
            'port_speed': self.get_port_speed,
            'interface_name': self.get_interface_name,
            'parent_interface_name': self.get_parent_interface_name,
            'parent_interface_media_type': self.get_parent_interface_media_type,
            'parent_interface_configured_speed': self.get_parent_interface_configured_speed,
            'parent_interface_port_speed': self.get_parent_interface_port_speed
        }

    # Dimensions
    def get_interface_name(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get('interface_name')

    def get_alias(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get('alias')

    def get_description(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get('description')

    def get_media_type(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get('media_type')

    def get_port_speed(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get('port_speed')

    def get_parent_interface_name(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get('parent_interface_name')

    def get_parent_interface_media_type(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get(
                'parent_interface_media_type')

    def get_parent_interface_port_speed(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get(
                'parent_interface_port_speed')

    def get_parent_interface_configured_speed(self, interface_index):
        return self.enrichment.get_enrichment_value('self', 'interface', interface_index).get(
                'parent_interface_configured_speed')

    # Metrics
    def get_bits_in(self, interface_index):
        if (ifHCInOctets + '.' + interface_index) in self._ifx_table_stats_map:
            return int(self._ifx_table_stats_map[ifHCInOctets + '.' + interface_index]) * 8
        else:
            return _MISSING_METRIC_VALUE

    def get_unicast_packets_in(self, interface_index):
        return int(self._ifx_table_stats_map.get(ifHCInUcastPkts + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_bits_out(self, interface_index):
        if (ifHCOutOctets + '.' + interface_index) in self._ifx_table_stats_map:
            return int(self._ifx_table_stats_map[ifHCOutOctets + '.' + interface_index]) * 8
        else:
            return _MISSING_METRIC_VALUE

    def get_unicast_packets_out(self, interface_index):
        return int(self._ifx_table_stats_map.get(ifHCOutUcastPkts + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_multicast_packets_in(self, interface_index):
        return int(self._ifx_table_stats_map.get(ifHCInMulticastPkts + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_multicast_packets_out(self, interface_index):
        return int(self._ifx_table_stats_map.get(ifHCOutMulticastPkts + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_broadcast_packets_in(self, interface_index):
        return int(self._ifx_table_stats_map.get(ifHCInBroadcastPkts + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_broadcast_packets_out(self, interface_index):
        return int(self._ifx_table_stats_map.get(ifHCOutBroadcastPkts + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_total_packets_in(self, interface_index):
        unicast_packets_in = self.get_unicast_packets_in(interface_index)
        multicast_packets_in = self.get_multicast_packets_in(interface_index)
        broadcast_packets_in = self.get_broadcast_packets_in(interface_index)

        if _MISSING_METRIC_VALUE not in [unicast_packets_in, multicast_packets_in, broadcast_packets_in]:
            return unicast_packets_in + multicast_packets_in + broadcast_packets_in
        else:
            return _MISSING_METRIC_VALUE

    def get_total_packets_out(self, interface_index):
        unicast_packets_out = self.get_unicast_packets_out(interface_index)
        multicast_packets_out = self.get_multicast_packets_out(interface_index)
        broadcast_packets_out = self.get_broadcast_packets_out(interface_index)

        if _MISSING_METRIC_VALUE not in [unicast_packets_out, multicast_packets_out, broadcast_packets_out]:
            return unicast_packets_out + multicast_packets_out + broadcast_packets_out
        else:
            return _MISSING_METRIC_VALUE

    def get_admin_state(self, interface_index):
        return int(self._if_table_stats_map.get(ifAdminStatus + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_oper_state(self, interface_index):
        return int(self._if_table_stats_map.get(ifOperStatus + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_oper_admin_state_mismatch(self, interface_index):
        return 0 if self.get_oper_state(interface_index) == self.get_admin_state(interface_index) else 1

    def get_discards_in(self, interface_index):
        return int(self._if_table_stats_map.get(ifInDiscards + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_errors_in(self, interface_index):
        return int(self._if_table_stats_map.get(ifInErrors + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_discards_out(self, interface_index):
        return int(self._if_table_stats_map.get(ifOutDiscards + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_errors_out(self, interface_index):
        return int(self._if_table_stats_map.get(ifOutErrors + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_mtu(self, interface_index):
        return int(self._if_table_stats_map.get(ifMtu + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_if_high_speed(self, interface_index):
        # n.b. adjusted value means I can't use 'get(..., _MISSING_METRIC_VALUE)' idiom
        # Mbps by definition
        return int(self._ifx_table_stats_map.get(ifHighSpeed + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_if_speed(self, interface_index):
        return int(self._if_table_stats_map.get(ifSpeed + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_configured_speed(self, index):
        return self.enrichment.get_enrichment_value('self', 'interface', index).get('configured_speed',
                                                                                    _MISSING_METRIC_VALUE)

    def get_errors_frame(self, interface_index):
        return int(self._dot3stats_map.get(dot3StatsAlignmentErrors + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_errors_crc(self, interface_index):
        return int(self._dot3stats_map.get(dot3StatsFCSErrors + '.' + interface_index, _MISSING_METRIC_VALUE))

    def get_errors_giants(self, interface_index):
        return int(self._dot3stats_map.get(dot3StatsFrameTooLongs + '.' + interface_index, _MISSING_METRIC_VALUE))

    @threaded_cached_property
    def interface_indices(self):
        result = set()
        for oid in self._ifx_table_stats_map:
            result.add(oid.split('.')[-1])
        return result

    def _getif_table_stats(self):
        result = dict()
        try:
            for oid in self._if_table_stats_map:
                index = oid.split('.')[-1]
                result[index] = dict()
                result[index]['admin_state'] = self.get_admin_state(index)
                result[index]['oper_state'] = self.get_oper_state(index)
                result[index]['oper_admin_state_mismatch'] = self.get_oper_admin_state_mismatch(index)
                result[index]['errors_in'] = self.get_errors_in(index)
                result[index]['errors_out'] = self.get_errors_out(index)
                result[index]['discards_in'] = self.get_discards_in(index)
                result[index]['discards_out'] = self.get_discards_out(index)
                result[index]['mtu'] = self.get_mtu(index)
            return result
        except Exception as e:
            self._polling_status.handle_exception('interface', e)

    def _getifx_table_stats(self):
        result = dict()
        try:
            for oid in self._ifx_table_stats_map:
                index = oid.split('.')[-1]
                result[index] = dict()
                result[index]['bits_in'] = self.get_bits_in(index)
                result[index]['bits_out'] = self.get_bits_out(index)
                result[index]['unicast_packets_in'] = self.get_unicast_packets_in(index)
                result[index]['unicast_packets_out'] = self.get_unicast_packets_out(index)
                result[index]['multicast_packets_in'] = self.get_multicast_packets_in(index)
                result[index]['multicast_packets_out'] = self.get_multicast_packets_out(index)
                result[index]['broadcast_packets_in'] = self.get_broadcast_packets_in(index)
                result[index]['broadcast_packets_out'] = self.get_broadcast_packets_out(index)
                result[index]['total_packets_in'] = self.get_total_packets_in(index)
                result[index]['total_packets_out'] = self.get_total_packets_out(index)
                result[index]['configured_speed'] = self.get_configured_speed(index)
            return result
        except Exception as e:
            self._polling_status.handle_exception('interface', e)

    def _getdot3stats(self):
        result = dict()
        try:
            # Use ifx_table_stats_map b/c dot3stats is not defined for every machine
            for oid in self._ifx_table_stats_map:
                index = oid.split('.')[-1]
                result[index] = dict()
                result[index]['errors_frame'] = self.get_errors_frame(index)
                result[index]['errors_crc'] = self.get_errors_crc(index)
                result[index]['errors_giants'] = self.get_errors_giants(index)
            return result
        except Exception as e:
            self._polling_status.handle_exception('interface', e)

    @staticmethod
    def _get_state_val(state):
        s = int(state)
        if s == 2:
            return _INTERFACE_STATES.DOWN
        elif s == 1:
            return _INTERFACE_STATES.UP
        else:
            return _INTERFACE_STATES.UNKNOWN

    def _build_ifx_table_stats_map(self):
        """Maps child oids of ifXTable to their respective values as PanoptesSNMPVariables"""
        ifx_table_stats = list()
        for metric in ifx_table_oids:
            for varbind in self._snmp_connection.bulk_walk(metric,
                                                           max_repetitions=self.snmp_configuration.max_repetitions):
                ifx_table_stats.append(varbind)

        self._ifx_table_stats_map = dict()
        for ent in ifx_table_stats:
            self._ifx_table_stats_map[ent.oid + '.' + ent.index] = ent.value

    def _build_if_table_stats_map(self):
        """Maps child oids of ifTable to their respective values as PanoptesSNMPVariables"""
        if_table_stats = list()
        for metric in if_table_oids:
            for varbind in self._snmp_connection.bulk_walk(metric,
                                                           max_repetitions=self.snmp_configuration.max_repetitions):
                if_table_stats.append(varbind)

        self._if_table_stats_map = dict()
        for ent in if_table_stats:
            self._if_table_stats_map[ent.oid + '.' + ent.index] = ent.value

    def _build_dot3stats_map(self):
        """Maps child oids of dot3statsTable to their respective values as PanoptesSNMPVariables"""
        dot3stats = list()
        for metric in dots3stats_table_oids:
            for varbind in self._snmp_connection.bulk_walk(metric,
                                                           max_repetitions=self.snmp_configuration.max_repetitions):
                dot3stats.append(varbind)

        self._dot3stats_map = dict()
        for ent in dot3stats:
            self._dot3stats_map[ent.oid + '.' + ent.index] = ent.value

    def _smart_add_dimension(self, method, dimension_name, index):
        dimension = method(index)
        if dimension is not None and PanoptesValidators.valid_nonempty_string(str(dimension)):
            self._interface_metrics_group.add_dimension(PanoptesMetricDimension(dimension_name, str(dimension)))
        else:
            self._interface_metrics_group.add_dimension(PanoptesMetricDimension(dimension_name,
                                                                                _DEFAULT_DIMENSION_VALUE))

    def get_results(self):
        self._polling_status = PanoptesPollingStatus(resource=self.resource,
                                                     execute_frequency=self.execute_frequency,
                                                     logger=self.logger,
                                                     metric_name='interface_polling_status')

        interface_metrics = dict()

        try:
            start_time = time.time()

            self._build_dot3stats_map()
            self._build_if_table_stats_map()
            self._build_ifx_table_stats_map()

            end_time = time.time()

            self._logger.info('SNMP calls for device %s completed in %.2f seconds' % (
                self.host, end_time - start_time))

            interface_metrics.update(self._getdot3stats())
            if_interface_metrics = self._getif_table_stats()
            ifx_interface_metrics = self._getifx_table_stats()

            # https://github.com/PyCQA/pylint/issues/1694
            for i in self.interface_indices:  # pylint: disable=E1133
                if i not in interface_metrics:
                    interface_metrics[i] = dict()
                interface_metrics[i].update(ifx_interface_metrics[i])
                interface_metrics[i].update(if_interface_metrics[i])

            for interface_index in interface_metrics.keys():
                self._interface_metrics_group = PanoptesMetricsGroup(self.resource, 'interface',
                                                                     self.execute_frequency)
                interface = interface_metrics[interface_index]

                for dimension_name, dimension_method in self._DIMENSION_MAP.items():
                    self._smart_add_dimension(method=dimension_method,
                                              dimension_name=dimension_name,
                                              index=interface_index
                                              )

                for metric in interface.keys():
                    metric_type = _METRIC_TYPE_MAP[metric]

                    if not isinstance(interface[metric], numbers.Number):
                        self._interface_metrics_group.add_metric(PanoptesMetric(str(metric),
                                                                                _MISSING_METRIC_VALUE,
                                                                                metric_type))
                    else:
                        self._interface_metrics_group.add_metric(PanoptesMetric(str(metric),
                                                                                interface[metric],
                                                                                metric_type))

                self._device_interface_metrics.add(self._interface_metrics_group)

            self._polling_status.handle_success('interface')
            self._logger.debug('Found interface metrics: "%s" for device "%s"' % (
                interface_metrics, self.host))
        except Exception as e:
            self._polling_status.handle_exception('interface', e)
        finally:
            self._device_interface_metrics.add(self._polling_status.device_status_metrics_group)
            return self._device_interface_metrics



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

The module implements the Enrichment Plugin Scheduler which finds any enrichment Plugins in the configured path and then
installs/updates a Celery Beat schedule for each enrichment plugin and configuration combination.

If the list or the configuration of the plugins changes, the schedule is updated

This module is expected to be imported and executed though the Celery 'beat' command line tool

Internally, there are two threads that get setup: one is the main thread that runs the Celery Beat service. The other is
the thread started by the Enrichment Plugin Scheduler to detect and update plugin/configuration changes
"""
import faulthandler
import sys
import time
from datetime import timedelta

from celery.signals import beat_init

from ..framework import const
from ..framework.exceptions import PanoptesBaseException
from ..framework.context import PanoptesContext
from ..framework.resources import PanoptesResourcesKeyValueStore, PanoptesResourceCache
from ..framework.celery_manager import PanoptesCeleryConfig
from ..framework.plugins.helpers import expires, time_limit
from ..framework.plugins.manager import PanoptesPluginManager
from ..framework.plugins.panoptes_base_plugin import PanoptesPluginConfigurationError
from ..framework.plugins.scheduler import PanoptesPluginScheduler
from ..framework.utilities.helpers import get_calling_module_name
from ..framework.utilities.key_value_store import PanoptesKeyValueStore
from .enrichment_plugin import PanoptesEnrichmentPlugin, PanoptesEnrichmentPluginInfo

panoptes_context = None
enrichment_plugin_scheduler = None
celery = None
logger = None


class PanoptesEnrichmentPluginSchedulerError(PanoptesBaseException):
    """
    The exception class for Enrichment Plugin Scheduler errors
    """
    pass


class PanoptesEnrichmentSchedulerKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for which uses the namespace demarcated for Enrichment Plugin Scheduler
    """

    def __init__(self, panoptes_context):
        super(PanoptesEnrichmentSchedulerKeyValueStore, self). \
            __init__(panoptes_context, const.ENRICHMENT_PLUGIN_SCHEDULER_KEY_VALUE_NAMESPACE)


class PanoptesEnrichmentPluginSchedulerContext(PanoptesContext):
    """
    This class implements a PanoptesContext with the following:
        A Key/Value Store for Enrichment Plugin Scheduler and Resource Manager
        A Zookeeper Client
    """

    def __init__(self):
        super(PanoptesEnrichmentPluginSchedulerContext, self).__init__(
                key_value_store_class_list=[PanoptesEnrichmentSchedulerKeyValueStore,
                                            PanoptesResourcesKeyValueStore],
                create_message_producer=False, create_zookeeper_client=True)


class PanoptesCeleryEnrichmentAgentConfig(PanoptesCeleryConfig):
    task_routes = {const.ENRICHMENT_PLUGIN_AGENT_MODULE_NAME: {'queue': const.ENRICHMENT_PLUGIN_AGENT_CELERY_APP_NAME}}

    def __init__(self):
        super(PanoptesCeleryEnrichmentAgentConfig, self). \
            __init__(app_name=const.ENRICHMENT_PLUGIN_SCHEDULER_CELERY_APP_NAME)


def enrichment_plugin_scheduler_task(celery_beat_service):
    """
    This function is the workhorse of the Enrichment Plugin Scheduler module. It detects changes in plugins and their
    configuration and updates the Celery Beat schedule accordingly.

    Args:
        celery_beat_service (celery.beat.Service): The Celery Beat Service instance associated with this Plugin\
        Scheduler

    Returns:
        None

    """
    start_time = time.time()

    try:
        resource_cache = PanoptesResourceCache(panoptes_context)
        resource_cache.setup_resource_cache()
    except:
        logger.exception('Could not create resource cache, skipping cycle')
        return

    try:
        plugin_manager = PanoptesPluginManager(plugin_type='enrichment',
                                               plugin_class=PanoptesEnrichmentPlugin,
                                               plugin_info_class=PanoptesEnrichmentPluginInfo,
                                               panoptes_context=panoptes_context,
                                               kv_store_class=PanoptesEnrichmentSchedulerKeyValueStore
                                               )
        plugins = plugin_manager.getPluginsOfCategory(category_name='enrichment')
        logger.info('Found %d plugins' % len(plugins))
    except:
        logger.exception('Error trying to load enrichment plugins, skipping cycle')
        return

    new_schedule = dict()

    for plugin in plugins:
        logger.info('Found plugin "%s", version %s at %s ' % (plugin.name, plugin.version, plugin.path))

        try:
            logger.info('Plugin "%s" has configuration: %s' % (plugin.name, plugin.config))
            logger.info('Plugin %s has plugin module time %s (UTC) and config mtime %s (UTC)' % (
                plugin.name, plugin.moduleMtime, plugin.configMtime))

            if plugin.execute_frequency <= 0:
                logger.info('Plugin %s has an invalid execution frequency (%d), skipping plugin' % (
                    plugin.name, plugin.execute_frequency))
                continue

            if not plugin.resource_filter:
                logger.info('Plugin "%s" does not have any resource filter specified, skipping plugin' %
                            plugin.name)
                continue
        except PanoptesPluginConfigurationError as e:
            logger.error('Error reading/parsing configuration for plugin "%s", skipping plugin. Error: %s' %
                         (plugin.name, repr(e)))

        try:
            resource_set = resource_cache.get_resources(plugin.resource_filter)
        except Exception as e:
            logger.info('Error in applying resource filter "%s" for plugin "%s", skipping plugin: %s' % (
                plugin.resource_filter, plugin.name, repr(e)))
            continue

        if len(resource_set) == 0:
            logger.info(
                    'No resources found for plugin "%s" after applying resource filter "%s", skipping plugin' % (
                        plugin.name, plugin.resource_filter))

        for resource in resource_set:
            logger.debug('Going to add task for plugin "%s" with execute frequency %d, args "%s", resources %s' % (
                plugin.name, plugin.execute_frequency, plugin.config, resource))

            task_name = ':'.join([plugin.normalized_name, plugin.signature, str(resource.resource_id)])

            new_schedule[task_name] = {
                'task': const.ENRICHMENT_PLUGIN_AGENT_MODULE_NAME,
                'schedule': timedelta(seconds=plugin.execute_frequency),
                'args': (plugin.name, resource.serialization_key),
                'options': {
                    'expires': expires(plugin),
                    'time_limit': time_limit(plugin)}
            }

    resource_cache.close_resource_cache()

    end_time = time.time()

    try:
        scheduler = celery_beat_service.scheduler
        scheduler.update(logger, new_schedule)
        logger.info('Scheduled %d tasks in %.2fs' % (len(new_schedule), end_time - start_time))
    except:
        logger.exception('Error in updating schedule for Polling Plugins')


def start_enrichment_plugin_scheduler():
    """
    The entry point for the Enrichment Plugin Scheduler

    This method creates a Panoptes Context and the Celery Instance for the Enrichment Plugin Scheduler

    Returns:
        None
    """
    global enrichment_plugin_scheduler, celery, logger, panoptes_context, resources_store

    try:
        panoptes_context = PanoptesEnrichmentPluginSchedulerContext()
    except Exception as e:
        sys.exit('Could not create a Panoptes Context: %s' % (repr(e)))

    try:
        celery_config = PanoptesCeleryEnrichmentAgentConfig()
    except Exception as e:
        sys.exit('Could not create a Celery Config object: %s' % repr(e))

    enrichment_plugin_scheduler = PanoptesPluginScheduler(
            panoptes_context=panoptes_context,
            plugin_type='enrichment',
            plugin_type_display_name='Enrichment',
            celery_config=celery_config,
            lock_timeout=const.ENRICHMENT_PLUGIN_SCHEDULER_LOCK_ACQUIRE_TIMEOUT,
            plugin_scheduler_task=enrichment_plugin_scheduler_task)

    logger = panoptes_context.logger
    celery = enrichment_plugin_scheduler.start()


@beat_init.connect
def celery_beat_service_started(sender=None, args=None, **kwargs):
    """
    This method is called after Celery Beat instantiates it's service

    Args:
        sender (celery.beat.Service): The Celery Beat Service which was started by Celery Beat
        args (dict): Arguments
        **kwargs (dict): Keyword Arguments

    Returns:
        None
    """
    global enrichment_plugin_scheduler
    sender.scheduler.panoptes_context = panoptes_context
    sender.scheduler.task_prefix = const.ENRICHMENT_PLUGIN_SCHEDULER_CELERY_TASK_PREFIX
    enrichment_plugin_scheduler.run(sender, args, **kwargs)


"""
This wrapper is to ensure that the Enrichment Plugin Scheduler only executes when called from Celery - prevents against
execution when imported from other modules (like Sphinx) or called from the command line
"""
if get_calling_module_name() == const.CELERY_LOADER_MODULE:
    faulthandler.enable()
    start_enrichment_plugin_scheduler()



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import time
import unittest
import json
import os

from mock import *

from yahoo_panoptes.framework.enrichment import PanoptesEnrichmentSet, PanoptesEnrichmentGroup, \
    PanoptesEnrichmentGroupSet, PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentEncoder, \
    PanoptesEnrichmentMultiGroupSet
from yahoo_panoptes.framework.resources import PanoptesResource
from yahoo_panoptes.enrichment.enrichment_plugin_agent import _store_enrichment_data, \
    PanoptesEnrichmentCacheKeyValueStore
from tests.test_framework import PanoptesMockRedis
from yahoo_panoptes.framework.context import PanoptesContext


mock_time = Mock()
mock_time.return_value = 1512629517.03121


def ordered(obj):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


def _get_test_conf_file():
    my_dir = os.path.dirname(os.path.realpath(__file__))
    panoptes_test_conf_file = os.path.join(my_dir, 'config_files/test_panoptes_config.ini')

    return my_dir, panoptes_test_conf_file


class PanoptesEnrichmentInterfaceSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                         'speed': {'type': 'integer'},
                         'index': {'type': 'integer'},
                         'status': {'type': 'string'}
                      }
        }
    }

    def __init__(self):
        super(PanoptesEnrichmentInterfaceSchemaValidator, self).__init__()


class PanoptesEnrichmentNeighborSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'vlan_id': {'type': 'integer', 'required': True},
                'property': {'type': 'string', 'required': True},
                'mac': {'type': 'string'}
            }
        }
    }

    def __init__(self):
        super(PanoptesEnrichmentNeighborSchemaValidator, self).__init__()


class TestEnrichmentFramework(unittest.TestCase):
    @patch('yahoo_panoptes.framework.resources.time', mock_time)
    def setUp(self):
        self.__panoptes_resource = PanoptesResource(resource_site='test', resource_class='test',
                                                    resource_subclass='test',
                                                    resource_type='test', resource_id='test', resource_endpoint='test',
                                                    resource_plugin='test')
        self.__panoptes_resource.add_metadata('test', 'test')

    def test_enrichment_set(self):
        enrichment_set = PanoptesEnrichmentSet('int_001')
        enrichment_set.add('speed', 1000)
        enrichment_set.add('index', 001)
        enrichment_set.add('status', 'up')
        self.assertEquals(enrichment_set.key, 'int_001')
        self.assertDictEqual(enrichment_set.value, {'status': 'up', 'index': 1, 'speed': 1000})
        self.assertEquals(len(enrichment_set), 3)

        enrichment_set1 = PanoptesEnrichmentSet('int_002', {'status': 'down', 'index': 2, 'speed': 1000})
        self.assertEquals(enrichment_set1.key, 'int_002')
        self.assertDictEqual(enrichment_set1.value, {'status': 'down', 'index': 2, 'speed': 1000})

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentSet('int_001', 'string')

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentSet('int_001', 100)

    def test_enrichment_schema_validator(self):
        validator = PanoptesEnrichmentInterfaceSchemaValidator()
        enrichment_set = PanoptesEnrichmentSet('int_001')
        enrichment_set.add('speed', 1000)
        enrichment_set.add('index', 001)
        enrichment_set.add('status', 'up')
        self.assertTrue(validator.validate(enrichment_set))

        enrichment_set.add('status', 01)
        self.assertFalse(validator.validate(enrichment_set))

    @patch('time.time', mock_time)
    def test_enrichment_group(self):
        interface_validation_object = PanoptesEnrichmentInterfaceSchemaValidator()
        neighbor_validation_object = PanoptesEnrichmentNeighborSchemaValidator()

        interface_data = \
            '''{"data": [
            {"int_001": {"index": 1, "speed": 1000, "status": "up"}},
            {"int_002": {"index": 2, "speed": 1000, "status": "down"}}],
            "metadata": {"_enrichment_group_creation_timestamp": %f, "_enrichment_ttl": 300, "_execute_frequency": 60},
            "namespace": "interface"}''' % mock_time.return_value

        neighbor_data = \
            '''{"data": [{"host_name": {"mac": "aa:bb:cc:dd:ee:ff", "property": "Test Property", "vlan_id": 501}}],
            "metadata": {"_enrichment_group_creation_timestamp": %f, "_enrichment_ttl": 600, "_execute_frequency": 120},
            "namespace": "neighbor"}''' % mock_time.return_value

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroup(1, interface_validation_object, 300, 60)

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroup('interface', 'non_validation_object', 300, 60)

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroup('interface', interface_validation_object, '300', 60)

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroup('interface', interface_validation_object, 300, '60')

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroup('interface', interface_validation_object, 0, 60)

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 0)

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60).\
                add_enrichment_set('not_PanoptesEnrichmentSet_obj')

        enrichment_set1 = PanoptesEnrichmentSet('int_001')
        enrichment_set1.add('speed', 1000)
        enrichment_set1.add('index', 001)
        enrichment_set1.add('status', 'up')

        enrichment_set2 = PanoptesEnrichmentSet('int_002')
        enrichment_set2.add('speed', 1000)
        enrichment_set2.add('index', 002)
        enrichment_set2.add('status', 'down')

        enrichment_group1 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)
        enrichment_group1.add_enrichment_set(enrichment_set1)
        enrichment_group1.add_enrichment_set(enrichment_set2)

        self.assertEqual(enrichment_group1.namespace, 'interface')
        self.assertEqual(enrichment_group1.enrichment_ttl, 300)
        self.assertEqual(enrichment_group1.execute_frequency, 60)
        self.assertEqual(enrichment_group1.enrichment_group_creation_timestamp, mock_time.return_value)
        self.assertEqual(ordered(json.loads(json.dumps(enrichment_group1.data, cls=PanoptesEnrichmentEncoder))),
                         ordered(json.loads(interface_data)['data']))
        self.assertEqual(ordered(json.loads(enrichment_group1.json())), ordered(json.loads(interface_data)))
        self.assertEquals(len(enrichment_group1), 2)

        enrichment_set3 = PanoptesEnrichmentSet('int_002')
        enrichment_set3.add('speed', 1000)
        enrichment_set3.add('index', 002)
        enrichment_set3.add('status', 'down')

        self.assertEqual(ordered(json.loads(enrichment_group1.json())), ordered(json.loads(interface_data)))
        self.assertEqual(ordered(enrichment_group1.metadata), ordered(json.loads(interface_data)['metadata']))
        self.assertEquals(len(enrichment_group1), 2)

        test_metadata = json.loads(interface_data)['metadata']
        test_metadata['metadata_key'] = 'metadata_value'

        enrichment_group1.upsert_metadata('metadata_key', 'metadata_value')
        self.assertEqual(ordered(enrichment_group1.metadata), ordered(test_metadata))
        enrichment_group1.upsert_metadata('ttl', 300)
        with self.assertRaises(ValueError):
            enrichment_group1.upsert_metadata('_enrichment_ttl', 300)
        with self.assertRaises(AssertionError):
            enrichment_group1.upsert_metadata('metadata', {})
        with self.assertRaises(AssertionError):
            enrichment_group1.upsert_metadata('metadata', [])

        enrichment_set4 = PanoptesEnrichmentSet('host_name')
        enrichment_set4.add('vlan_id', 501)
        enrichment_set4.add('property', 'Test Property')
        enrichment_set4.add('mac', 'aa:bb:cc:dd:ee:ff')

        enrichment_group2 = PanoptesEnrichmentGroup('neighbor', neighbor_validation_object, 600, 120)
        enrichment_group2.add_enrichment_set(enrichment_set4)

        self.assertEqual(ordered(json.loads(enrichment_group2.json())), ordered(json.loads(neighbor_data)))
        self.assertEquals(len(enrichment_group2), 1)

        enrichment_set5 = PanoptesEnrichmentSet('host_name01')
        enrichment_set5.add('vlan_id', 502)
        enrichment_set5.add('property', 'Netops01.US')

        enrichment_set6 = PanoptesEnrichmentSet('host_name02')
        enrichment_set6.add('vlan_id', 503)
        enrichment_set6.add('mac', 'aa:bb:cc:dd:ee:ff')

        enrichment_group3 = PanoptesEnrichmentGroup('neighbor', neighbor_validation_object, 600, 120)
        enrichment_group3.add_enrichment_set(enrichment_set5)
        with self.assertRaises(AssertionError):
            enrichment_group3.add_enrichment_set(enrichment_set6)

        interface_store_data = '{"int_001": {"index": 1, "speed": 1000, "status": "up"}, ' \
                               '"int_002": {"index": 2, "speed": 1000, "status": "down"}}'

        neighbor_store_data = '{"host_name": {"mac": "aa:bb:cc:dd:ee:ff", "property": "Test Property", "vlan_id": 501}}'

        self.assertEquals(ordered(json.loads(enrichment_group1.serialize_data())),
                          ordered(json.loads(interface_store_data)))
        self.assertEquals(ordered(json.loads(enrichment_group2.serialize_data())),
                          ordered(json.loads(neighbor_store_data)))

        enrichment_group1.upsert_metadata('ttl', 300)
        with self.assertRaises(ValueError):
            enrichment_group1.upsert_metadata('_enrichment_ttl', 300)

        interface_data_serialized = '''{{"data": {{"int_001": {{"index": 1, "speed": 1000, "status": "up"}},
        "int_002": {{"index": 2, "speed": 1000, "status": "down"}}}}, "metadata":
        {{"_enrichment_group_creation_timestamp": {:.5f}, "_enrichment_ttl": 300, "_execute_frequency": 60,
        "metadata_key": "metadata_value", "ttl": 300}}}}'''.format(mock_time.return_value)

        neighbor_data_serialized = '''{{"data": {{"host_name": {{"mac": "aa:bb:cc:dd:ee:ff", "property": "Test Property",
        "vlan_id": 501}}}}, "metadata": {{"_enrichment_group_creation_timestamp": {:.5f},
        "_enrichment_ttl": 600, "_execute_frequency": 120}}}}'''.format(mock_time.return_value)

        self.assertEquals(ordered(json.loads(enrichment_group1.serialize())),
                          ordered(json.loads(interface_data_serialized)))

        self.assertEquals(ordered(json.loads(enrichment_group2.serialize())),
                          ordered(json.loads(neighbor_data_serialized)))

    @patch('time.time', mock_time)
    def test_enrichment_group_set(self):
        interface_validation_object = PanoptesEnrichmentInterfaceSchemaValidator()
        neighbor_validation_object = PanoptesEnrichmentNeighborSchemaValidator()

        panoptes_resource = self.__panoptes_resource

        enrichment_data = \
            '''{{"enrichment": [{{"metadata": {{"_enrichment_group_creation_timestamp": {:.5f}, "_enrichment_ttl": 600,
            "_execute_frequency": 120}}, "data": [{{"host_name":
            {{"mac": "aa:bb:cc:dd:ee:ff", "property": "Test Property", "vlan_id": 501}}}}],
            "namespace": "neighbor"}}, {{"metadata": {{"_enrichment_group_creation_timestamp": {:.5f},
            "_enrichment_ttl": 300,
            "_execute_frequency": 60}}, "data": [
            {{"int_001": {{"index": 1, "speed": 1000, "status": "up"}}}}, {{"int_002": {{"index": 2, "speed": 1000,
            "status": "down"}}}}], "namespace": "interface"}}],
            "enrichment_group_set_creation_timestamp": {:.5f}, "resource": {{"resource_class": "test",
            "resource_creation_timestamp": {:.5f}, "resource_endpoint": "test", "resource_id": "test",
            "resource_metadata": {{"_resource_ttl": "604800", "test": "test"}}, "resource_plugin": "test",
            "resource_site": "test",
            "resource_subclass": "test", "resource_type": "test"}}}}'''.format(mock_time.return_value,
                                                                               mock_time.return_value,
                                                                               mock_time.return_value,
                                                                               mock_time.return_value)

        enrichment_set1 = PanoptesEnrichmentSet('int_001')
        enrichment_set1.add('speed', 1000)
        enrichment_set1.add('index', 001)
        enrichment_set1.add('status', 'up')

        enrichment_set2 = PanoptesEnrichmentSet('int_002')
        enrichment_set2.add('speed', 1000)
        enrichment_set2.add('index', 002)
        enrichment_set2.add('status', 'down')

        enrichment_group1 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)
        enrichment_group1.add_enrichment_set(enrichment_set1)
        enrichment_group1.add_enrichment_set(enrichment_set2)

        enrichment_set3 = PanoptesEnrichmentSet('host_name')
        enrichment_set3.add('vlan_id', 501)
        enrichment_set3.add('property', 'Test Property')
        enrichment_set3.add('mac', 'aa:bb:cc:dd:ee:ff')

        enrichment_group2 = PanoptesEnrichmentGroup('neighbor', neighbor_validation_object, 600, 120)
        enrichment_group2.add_enrichment_set(enrichment_set3)

        enrichment_group_set1 = PanoptesEnrichmentGroupSet(panoptes_resource)
        enrichment_group_set1.add_enrichment_group(enrichment_group1)
        enrichment_group_set1.add_enrichment_group(enrichment_group2)
        self.assertEquals(len(enrichment_group_set1), 2)
        group_set_repr = "PanoptesEnrichmentGroupSet({{'enrichment_group_set_creation_timestamp': {:.5f}, " \
                         "'resource': plugin|test|site|test|class|test|subclass|test" \
                         "|type|test|id|test|endpoint|test, " \
                         "'enrichment': set([PanoptesEnrichmentGroup({{'data': set([PanoptesEnrichmentSet({{" \
                         "'int_001': {{'status': 'up', 'index': 1, 'speed': 1000}}}}), " \
                         "PanoptesEnrichmentSet({{'int_002': {{'status': 'down', 'index': 2, 'speed': 1000}}}})]), " \
                         "'namespace': 'interface', 'metadata': " \
                         "{{'_enrichment_group_creation_timestamp': {:.5f}, '_enrichment_ttl': 300, " \
                         "'_execute_frequency': 60}}}}), PanoptesEnrichmentGroup({{" \
                         "'data': set([PanoptesEnrichmentSet(" \
                         "{{'host_name': {{" \
                         "'mac': 'aa:bb:cc:dd:ee:ff', 'property': 'Test Property', 'vlan_id': 501}}}})]), " \
                         "'namespace': 'neighbor', 'metadata': {{'_enrichment_group_creation_timestamp': " \
                         "{:.5f}, '_enrichment_ttl': 600, " \
                         "'_execute_frequency': 120}}}})])}})".format(mock_time.return_value,
                                                                      mock_time.return_value,
                                                                      mock_time.return_value)
        self.assertEquals(repr(enrichment_group_set1), group_set_repr)

        self.assertIsInstance(enrichment_group_set1.resource, PanoptesResource)
        self.assertEqual(enrichment_group_set1.enrichment_group_set_creation_timestamp, mock_time.return_value)

        self.assertEqual(
            ordered(json.loads(json.dumps(enrichment_group_set1.enrichment, cls=PanoptesEnrichmentEncoder))),
            ordered(json.loads(enrichment_data)['enrichment']))

        self.assertEqual(ordered(json.loads(enrichment_group_set1.json())['enrichment']),
                         ordered(json.loads(enrichment_data)['enrichment']))

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroupSet('bad_resource')

        with self.assertRaises(AssertionError):
            PanoptesEnrichmentGroupSet(panoptes_resource).add_enrichment_group('non_PanoptesEnrichmentGroup_obj')

        enrichment_group_set2 = PanoptesEnrichmentGroupSet(panoptes_resource)
        enrichment_group3 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)

        with self.assertRaises(AssertionError):
            enrichment_group_set2.add_enrichment_group(enrichment_group3)

        self.assertFalse(enrichment_group_set1 == enrichment_group1)
        self.assertFalse(enrichment_group_set1 == enrichment_group_set2)

    @patch('time.time', mock_time)
    @patch('yahoo_panoptes.framework.resources.time', mock_time)
    def test_multi_enrichment_group_set(self):
        interface_validation_object = PanoptesEnrichmentInterfaceSchemaValidator()
        neighbor_validation_object = PanoptesEnrichmentNeighborSchemaValidator()

        panoptes_resource = self.__panoptes_resource

        multi_enrichment_results_data = \
            {
                "group_sets": [
                    {
                        "enrichment": [
                            {
                                "data": [
                                    {
                                        "host_name": {
                                            "mac": "aa:bb:cc:dd:ee:ff",
                                            "property": "Test Property",
                                            "vlan_id": 501
                                        }
                                    },
                                    {
                                        "host_name01": {
                                            "mac": "aa:bb:cc:dd:ee:ff",
                                            "property": "Test Property",
                                            "vlan_id": 502
                                        }
                                    }
                                ],
                                "metadata": {
                                    "_enrichment_group_creation_timestamp": mock_time.return_value,
                                    "_enrichment_ttl": 600,
                                    "_execute_frequency": 120
                                },
                                "namespace": "neighbor"
                            },
                            {
                                "data": [
                                    {
                                        "int_001": {
                                            "index": 1,
                                            "speed": 1000,
                                            "status": "up"
                                        }
                                    }
                                ],
                                "metadata": {
                                    "_enrichment_group_creation_timestamp": mock_time.return_value,
                                    "_enrichment_ttl": 300,
                                    "_execute_frequency": 60
                                },
                                "namespace": "interface"
                            }
                        ],
                        "enrichment_group_set_creation_timestamp": mock_time.return_value,
                        "resource": {
                            "resource_class": "test_class",
                            "resource_creation_timestamp": mock_time.return_value,
                            "resource_endpoint": "test_endpoint01",
                            "resource_id": "test_resource_id01",
                            "resource_metadata": {
                                "_resource_ttl": "604800"
                            },
                            "resource_plugin": "test_plugin",
                            "resource_site": "test_site",
                            "resource_subclass": "test_subclass",
                            "resource_type": "test_type"
                        }
                    },
                    {
                        "enrichment": [
                            {
                                "data": [
                                    {
                                        "int_001": {
                                            "index": 1,
                                            "speed": 1000,
                                            "status": "up"
                                        }
                                    }
                                ],
                                "metadata": {
                                    "_enrichment_group_creation_timestamp": mock_time.return_value,
                                    "_enrichment_ttl": 300,
                                    "_execute_frequency": 60
                                },
                                "namespace": "interface"
                            },
                            {
                                "data": [
                                    {
                                        "host_name": {
                                            "mac": "aa:bb:cc:dd:ee:ff",
                                            "property": "Test Property",
                                            "vlan_id": 501
                                        }
                                    }
                                ],
                                "metadata": {
                                    "_enrichment_group_creation_timestamp": mock_time.return_value,
                                    "_enrichment_ttl": 600,
                                    "_execute_frequency": 120
                                },
                                "namespace": "neighbor"
                            }
                        ],
                        "enrichment_group_set_creation_timestamp": mock_time.return_value,
                        "resource": {
                            "resource_class": "test",
                            "resource_creation_timestamp": mock_time.return_value,
                            "resource_endpoint": "test",
                            "resource_id": "test",
                            "resource_metadata": {
                                "_resource_ttl": "604800",
                                "test": "test"
                            },
                            "resource_plugin": "test",
                            "resource_site": "test",
                            "resource_subclass": "test",
                            "resource_type": "test"
                        }
                    }
                ]
            }

        enrichment_set1 = PanoptesEnrichmentSet('int_001')
        enrichment_set1.add('speed', 1000)
        enrichment_set1.add('index', 001)
        enrichment_set1.add('status', 'up')

        enrichment_group1 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)
        enrichment_group1.add_enrichment_set(enrichment_set1)

        enrichment_set3 = PanoptesEnrichmentSet('host_name')
        enrichment_set3.add('vlan_id', 501)
        enrichment_set3.add('property', 'Test Property')
        enrichment_set3.add('mac', 'aa:bb:cc:dd:ee:ff')

        enrichment_group2 = PanoptesEnrichmentGroup('neighbor', neighbor_validation_object, 600, 120)
        enrichment_group2.add_enrichment_set(enrichment_set3)

        enrichment_group_set1 = PanoptesEnrichmentGroupSet(panoptes_resource)
        enrichment_group_set1.add_enrichment_group(enrichment_group1)
        enrichment_group_set1.add_enrichment_group(enrichment_group2)

        panoptes_resource01 = PanoptesResource(resource_site='test_site',
                                               resource_class='test_class',
                                               resource_subclass='test_subclass',
                                               resource_type='test_type',
                                               resource_id='test_resource_id01',
                                               resource_endpoint='test_endpoint01',
                                               resource_plugin='test_plugin')

        panoptes_resource02 = PanoptesResource(resource_site='test_site',
                                               resource_class='test_class',
                                               resource_subclass='test_subclass',
                                               resource_type='test_type',
                                               resource_id='test_resource_id02',
                                               resource_endpoint='test_endpoint02',
                                               resource_plugin='test_plugin')

        enrichment_set4 = PanoptesEnrichmentSet('host_name01')
        enrichment_set4.add('vlan_id', 502)
        enrichment_set4.add('property', 'Test Property')
        enrichment_set4.add('mac', 'aa:bb:cc:dd:ee:ff')

        enrichment_group3 = PanoptesEnrichmentGroup('neighbor', neighbor_validation_object, 600, 120)
        enrichment_group3.add_enrichment_set(enrichment_set3)
        enrichment_group3.add_enrichment_set(enrichment_set4)

        enrichment_group_set2 = PanoptesEnrichmentGroupSet(panoptes_resource01)
        enrichment_group_set2.add_enrichment_group(enrichment_group1)
        enrichment_group_set2.add_enrichment_group(enrichment_group3)

        multi_enrichment_group_set = PanoptesEnrichmentMultiGroupSet()
        multi_enrichment_group_set.add_enrichment_group_set(enrichment_group_set1)
        multi_enrichment_group_set.add_enrichment_group_set(enrichment_group_set2)

        multi_enrichment_group_set_repr = \
            "PanoptesEnrichmentMultiGroupSet({{'group_sets': set([PanoptesEnrichmentGroupSet({{" \
            "'enrichment_group_set_creation_timestamp': {:.5f}, 'resource': plugin|test_plugin|" \
            "site|test_site|class|test_class|subclass|test_subclass|type|test_type|id|test_resource_id01|" \
            "endpoint|test_endpoint01, 'enrichment': set([PanoptesEnrichmentGroup({{'data': set([" \
            "PanoptesEnrichmentSet({{'int_001': {{'status': 'up', 'index': 1, 'speed': 1000}}}})]), " \
            "'namespace': 'interface', 'metadata': {{'_enrichment_group_creation_timestamp': {:.5f}, " \
            "'_enrichment_ttl': 300, '_execute_frequency': 60}}}}), PanoptesEnrichmentGroup({{" \
            "'data': set([PanoptesEnrichmentSet({{'host_name': {{" \
            "'mac': 'aa:bb:cc:dd:ee:ff', 'property': 'Test Property', " \
            "'vlan_id': 501}}}}), PanoptesEnrichmentSet({{'host_name01': {{'mac': 'aa:bb:cc:dd:ee:ff', " \
            "'property': 'Test Property', 'vlan_id': 502}}}})]), 'namespace': 'neighbor', 'metadata': {{" \
            "'_enrichment_group_creation_timestamp': {:.5f}, '_enrichment_ttl': 600, " \
            "'_execute_frequency': 120}}}})])}}), PanoptesEnrichmentGroupSet({{" \
            "'enrichment_group_set_creation_timestamp': {:.5f}, 'resource': plugin|test|site|test|" \
            "class|test|subclass|test|type|test|id|test|endpoint|test, 'enrichment': set([PanoptesEnrichmentGroup({{" \
            "'data': set([PanoptesEnrichmentSet({{'int_001': {{'status': 'up', 'index': 1, 'speed': 1000}}}})]), " \
            "'namespace': 'interface', 'metadata': {{'_enrichment_group_creation_timestamp': {:.5f}, " \
            "'_enrichment_ttl': 300, '_execute_frequency': 60}}}}), PanoptesEnrichmentGroup({{'data': set([" \
            "PanoptesEnrichmentSet({{'host_name': {{'mac': 'aa:bb:cc:dd:ee:ff', 'property': 'Test Property', " \
            "'vlan_id': 501}}}})]), 'namespace': 'neighbor', 'metadata': {{'_enrichment_group_creation_timestamp': " \
            "{:.5f}, '_enrichment_ttl': 600, '_execute_frequency': 120}}}})])}})])}})".format(
                mock_time.return_value,
                mock_time.return_value,
                mock_time.return_value,
                mock_time.return_value,
                mock_time.return_value,
                mock_time.return_value)

        self.assertEquals(repr(multi_enrichment_group_set), multi_enrichment_group_set_repr)

        self.assertEquals(len(multi_enrichment_group_set.enrichment_group_sets), 2)

        self.assertEquals(ordered(json.loads(multi_enrichment_group_set.json())),
                          ordered(multi_enrichment_results_data))

        self.assertEquals(len(multi_enrichment_group_set), 2)

        multi_enrichment_group_set.add_enrichment_group_set(enrichment_group_set2)
        self.assertEquals(len(multi_enrichment_group_set), 2)

        enrichment_group_set3 = PanoptesEnrichmentGroupSet(panoptes_resource02)
        enrichment_group_set3.add_enrichment_group(enrichment_group1)
        multi_enrichment_group_set.add_enrichment_group_set(enrichment_group_set3)
        self.assertEquals(len(multi_enrichment_group_set), 3)

        with self.assertRaises(AssertionError):
            multi_enrichment_group_set.add_enrichment_group_set('non_enrichment_group')

        enrichment_group_set3 = PanoptesEnrichmentGroupSet(panoptes_resource01)
        with self.assertRaises(AssertionError):
            multi_enrichment_group_set.add_enrichment_group_set(enrichment_group_set3)


class TestPanoptesEnrichmentCacheStore(unittest.TestCase):
    @patch('redis.StrictRedis', PanoptesMockRedis)
    @patch('time.time', mock_time)
    def setUp(self):
        self.my_dir, self.panoptes_test_conf_file = _get_test_conf_file()
        self._panoptes_context = PanoptesContext(self.panoptes_test_conf_file,
                                                 key_value_store_class_list=[PanoptesEnrichmentCacheKeyValueStore])

        self._enrichment_kv = self._panoptes_context.get_kv_store(PanoptesEnrichmentCacheKeyValueStore)

        self._panoptes_resource = PanoptesResource(resource_site='test_site',
                                                   resource_class='test_class',
                                                   resource_subclass='test_subclass',
                                                   resource_type='test_type',
                                                   resource_id='test_resource_id',
                                                   resource_endpoint='test_endpoint',
                                                   resource_plugin='test_plugin')

        interface_validation_object = PanoptesEnrichmentInterfaceSchemaValidator()
        neighbor_validation_object = PanoptesEnrichmentNeighborSchemaValidator()

        enrichment_set1 = PanoptesEnrichmentSet('int_001')
        enrichment_set1.add('speed', 1000)
        enrichment_set1.add('index', 001)
        enrichment_set1.add('status', 'up')

        enrichment_set2 = PanoptesEnrichmentSet('int_002')
        enrichment_set2.add('speed', 1000)
        enrichment_set2.add('index', 002)
        enrichment_set2.add('status', 'down')

        enrichment_group1 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)
        enrichment_group1.add_enrichment_set(enrichment_set1)
        enrichment_group1.add_enrichment_set(enrichment_set2)

        enrichment_set3 = PanoptesEnrichmentSet('host_name')
        enrichment_set3.add('vlan_id', 501)
        enrichment_set3.add('property', 'Test Property')
        enrichment_set3.add('mac', 'aa:bb:cc:dd:ee:ff')

        enrichment_group2 = PanoptesEnrichmentGroup('neighbor', neighbor_validation_object, 600, 120)
        enrichment_group2.add_enrichment_set(enrichment_set3)

        self.enrichment_group_set1 = PanoptesEnrichmentGroupSet(self._panoptes_resource)
        self.enrichment_group_set1.add_enrichment_group(enrichment_group1)
        self.enrichment_group_set1.add_enrichment_group(enrichment_group2)

        self._panoptes_resource01 = PanoptesResource(resource_site='test_site',
                                                     resource_class='test_class',
                                                     resource_subclass='test_subclass',
                                                     resource_type='test_type',
                                                     resource_id='test_resource_id01',
                                                     resource_endpoint='test_endpoint01',
                                                     resource_plugin='test_plugin')

        enrichment_set4 = PanoptesEnrichmentSet('host_name01')
        enrichment_set4.add('vlan_id', 502)
        enrichment_set4.add('property', 'Test Property')
        enrichment_set4.add('mac', 'aa:bb:cc:dd:ee:ff')

        enrichment_group3 = PanoptesEnrichmentGroup('neighbor', neighbor_validation_object, 600, 120)
        enrichment_group3.add_enrichment_set(enrichment_set3)
        enrichment_group3.add_enrichment_set(enrichment_set4)

        self.enrichment_group_set2 = PanoptesEnrichmentGroupSet(self._panoptes_resource01)
        self.enrichment_group_set2.add_enrichment_group(enrichment_group1)
        self.enrichment_group_set2.add_enrichment_group(enrichment_group3)

        self._multi_enrichment_group_set = PanoptesEnrichmentMultiGroupSet()
        self._multi_enrichment_group_set.add_enrichment_group_set(self.enrichment_group_set1)
        self._multi_enrichment_group_set.add_enrichment_group_set(self.enrichment_group_set2)

    @patch('time.time', mock_time)
    def test_panoptes_enrichment_set(self):
        enrichment_set1 = PanoptesEnrichmentSet('int_001')
        enrichment_set1.add('speed', 1000)
        enrichment_set1.add('index', 001)
        enrichment_set1.add('status', 'up')
        self.assertEquals(enrichment_set1.json(),
                          '{"int_001": {"index": 1, "speed": 1000, "status": "up"}}')
        self.assertEquals(repr(enrichment_set1),
                          "PanoptesEnrichmentSet({'int_001': {'status': 'up', 'index': 1, 'speed': 1000}})")

        enrichment_set2 = PanoptesEnrichmentSet('int_002')
        enrichment_set2.add('speed', 1000)
        enrichment_set2.add('index', 002)
        enrichment_set2.add('status', 'down')

        interface_validation_object = PanoptesEnrichmentInterfaceSchemaValidator()
        enrichment_group1 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)

        self.assertFalse(enrichment_set1 == enrichment_group1)
        self.assertFalse(enrichment_set1 == enrichment_set2)

    @patch('time.time', mock_time)
    def test_panoptes_enrichment_group(self):
        interface_validation_object = PanoptesEnrichmentInterfaceSchemaValidator()
        enrichment_group1 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)
        self.assertEquals(enrichment_group1.enrichment_schema, PanoptesEnrichmentInterfaceSchemaValidator.schema)
        self.assertEquals(repr(enrichment_group1), "PanoptesEnrichmentGroup({{'data': set([]), "
                                                   "'namespace': 'interface', "
                                                   "'metadata': {{"
                                                   "'_enrichment_group_creation_timestamp': {:.5f}, "
                                                   "'_enrichment_ttl': 300, "
                                                   "'_execute_frequency': 60}}}})".format(mock_time.return_value))

        enrichment_set1 = PanoptesEnrichmentSet('int_001')
        enrichment_set1.add('speed', 1000)
        enrichment_set1.add('index', 001)
        enrichment_set1.add('status', 'up')

        self.assertFalse(enrichment_group1 == enrichment_set1)

        enrichment_group2 = PanoptesEnrichmentGroup('interface', interface_validation_object, 300, 60)
        enrichment_group3 = PanoptesEnrichmentGroup('other_namespace', interface_validation_object, 300, 60)
        self.assertTrue(enrichment_group1 == enrichment_group2)
        self.assertFalse(enrichment_group1 == enrichment_group3)

    @patch('time.time', mock_time)
    def test_store_enrichment_data_enrichment_group_set(self):
        interface_result_data = \
            {
                "data": {
                    "int_001": {
                        "index": 1,
                        "speed": 1000,
                        "status": "up"
                    },
                    "int_002": {
                        "index": 2,
                        "speed": 1000,
                        "status": "down"
                    }
                },
                "metadata": {
                    "_enrichment_group_creation_timestamp": mock_time.return_value,
                    "_enrichment_ttl": 300,
                    "_execute_frequency": 60
                }
            }

        neighbor_result_data = \
            {
                "data": {
                    "host_name": {
                        "mac": "aa:bb:cc:dd:ee:ff",
                        "property": "Test Property",
                        "vlan_id": 501
                    }
                },
                "metadata": {
                    "_enrichment_group_creation_timestamp": mock_time.return_value,
                    "_enrichment_ttl": 600,
                    "_execute_frequency": 120
                }
            }

        _store_enrichment_data(self._panoptes_context, self.enrichment_group_set1, 'PanoptesPluginInfo')

        self.assertNotEquals(ordered(interface_result_data),
                             ordered(json.loads(self._enrichment_kv.get('test_resource_id:neighbor'))))

        self.assertEquals(ordered(interface_result_data),
                          ordered(json.loads(self._enrichment_kv.get('test_resource_id:interface'))))

        self.assertEquals(ordered(neighbor_result_data),
                          ordered(json.loads(self._enrichment_kv.get('test_resource_id:neighbor'))))

    def test_store_enrichment_data_enrichment_multi_group_set(self):

        enrichment_result_keys = ['test_resource_id01:interface', 'test_resource_id01:neighbor',
                                  'test_resource_id:interface', 'test_resource_id:neighbor']

        neighbor_result_data = \
            {"data": {"host_name": {"mac": "aa:bb:cc:dd:ee:ff", "property": "Test Property", "vlan_id": 501},
                      "host_name01": {"mac": "aa:bb:cc:dd:ee:ff", "property": "Test Property", "vlan_id": 502}},
             "metadata": {"_enrichment_group_creation_timestamp": mock_time.return_value, "_enrichment_ttl": 600,
                          "_execute_frequency": 120}}

        _store_enrichment_data(self._panoptes_context, self._multi_enrichment_group_set, 'PanoptesPluginInfo')

        self.assertEquals(enrichment_result_keys, self._enrichment_kv.find_keys('*'))

        self.assertEquals(ordered(neighbor_result_data),
                          ordered(json.loads(self._enrichment_kv.get('test_resource_id01:neighbor'))))

        self.assertNotEquals(ordered(neighbor_result_data),
                             ordered(json.loads(self._enrichment_kv.get('test_resource_id01:interface'))))



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.
"""

import unittest

from yahoo_panoptes.framework.utilities.helpers import *


class TestHelpers(unittest.TestCase):
    def test_normalize_plugin_name(self):
        self.assertEqual(normalize_plugin_name('Test Plugin Name'), 'Test_Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test/Plugin/Name'), 'Test_Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test/Plugin Name'), 'Test_Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test.Plugin.Name'), 'Test_Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test Plugin.Name'), 'Test_Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test/Plugin.Name'), 'Test_Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test_Plugin.Name'), 'Test__Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test.Plugin_Name'), 'Test_Plugin__Name')
        self.assertEqual(normalize_plugin_name('Test_Plugin_Name'), 'Test__Plugin__Name')
        self.assertEqual(normalize_plugin_name('Test!_Plugin.Name'), 'Test___Plugin_Name')
        self.assertEqual(normalize_plugin_name('Test.Plugin!#Name'), 'Test_Plugin__Name')
        with self.assertRaises(AssertionError):
            normalize_plugin_name(None)
            normalize_plugin_name(1)

    def test_get_module_mtime(self):
        self.assertGreater(get_module_mtime('.'), 0)
        my_module_path = os.path.splitext(os.path.abspath(__file__))[0]
        self.assertGreater(get_module_mtime(my_module_path), 0)
        self.assertEqual(get_module_mtime('/none/existent/module/path'), 0)
        with self.assertRaises(AssertionError):
            get_module_mtime(None)
            get_module_mtime(1)

    def test_get_hostname(self):
        self.assertIsInstance(get_hostname(), str)

    def test_resolve_hostnames(self):
        self.assertEqual(resolve_hostnames(['localhost'], 1), [('localhost', '127.0.0.1')])
        self.assertEqual(resolve_hostnames(['non.existent.host'], 1), [('non.existent.host', None)])
        with self.assertRaises(AssertionError):
            resolve_hostnames(None, 1)
        with self.assertRaises(AssertionError):
            resolve_hostnames(['localhost'], '1')

    def test_get_od_tid(self):
        self.assertGreater(get_os_tid(), 0)

    def test_get_calling_module_name(self):
        self.assertRegexpMatches(get_calling_module_name(1), '.*test_helpers$')
        with self.assertRaises(AssertionError):
            get_calling_module_name('1')

    def test_convert_kv_str_to_dict(self):
        with self.assertRaises(AssertionError):
            convert_kv_str_to_dict(kv=None, prefix='prefix')

        with self.assertRaises(AssertionError):
            convert_kv_str_to_dict(kv='key|value', prefix=None)

        with self.assertRaises(AssertionError):
            convert_kv_str_to_dict(kv='key|value', prefix='prefix', kv_delimiter=None)

        with self.assertRaises(AssertionError):
            convert_kv_str_to_dict(kv='key|value', prefix='prefix', kv_delimiter='|', prefix_delimiter=None)

        self.assertEquals(convert_kv_str_to_dict(
                kv='metro|WA State|region|Western US|continent|North America|latitude|47.615|longitude|-122.339',
                prefix='geo'), {'geo_metro': 'WA State', 'geo_region': 'Western US', 'geo_continent': 'North America',
                                'geo_latitude': '47.615', 'geo_longitude': '-122.339'})

    def test_transform_index_ipv6_address(self):
        self.assertEquals(transform_index_ipv6_address('32.1.73.152.0.88.206.3.0.0.0.0.0.0.0.1'),
                          '2001:4998:58:ce03::1')
        self.assertEquals(transform_index_ipv6_address('254.128.0.0.0.0.0.0.0.5.115.255.254.160.0.3'),
                          'fe80::5:73ff:fea0:3')

    def test_transform_octet_to_mac(self):
        self.assertEquals(transform_octet_to_mac(u'\xe4\xc7"\xdbJ\x08'), 'E4:C7:22:DB:4A:08')
        self.assertEquals(transform_octet_to_mac(u'\xe4\xc7"\xdbJ\t'), 'E4:C7:22:DB:4A:09')


if __name__ == '__main__':
    unittest.main()



import copy
import json
import logging
import os
import subprocess

from mock import Mock, patch, create_autospec
from yahoo_panoptes.framework.context import PanoptesContext
from yahoo_panoptes.framework.enrichment import PanoptesEnrichmentCache, PanoptesEnrichmentCacheKeyValueStore
from yahoo_panoptes.framework.plugins.panoptes_base_plugin import PanoptesPluginRuntimeError
from yahoo_panoptes.framework.plugins.context import PanoptesPluginWithEnrichmentContext, PanoptesPluginContext
from yahoo_panoptes.framework.resources import PanoptesResource
from yahoo_panoptes.framework.utilities.helpers import ordered
from yahoo_panoptes.framework.utilities.secrets import PanoptesSecretsStore

from tests.mock_redis import PanoptesMockRedis

mock_time = Mock()
mock_time.return_value = 1512629517.03121

snmp_simulator = None


def setup_module_default(plugin_pwd, snmp_sim_listen='127.0.0.1:10161', snmp_sim_data_dir='data/recording'):
    global snmp_simulator

    snmp_sim_data_dir = os.path.join(plugin_pwd, snmp_sim_data_dir)

    try:
        snmp_simulator = subprocess.Popen(
                ['snmpsimd.py',
                 '--data-dir={}'.format(snmp_sim_data_dir),
                 '--agent-udpv4-endpoint={}'.format(snmp_sim_listen),
                 '--logging-method=null'],
        )
    except Exception as e:
        raise IOError('Failed to start SNMP simulator: {}'.format(repr(e)))


def tear_down_module_default():
    try:
        if snmp_simulator is not None:
            snmp_simulator.kill()
    except OSError:
        pass


class DiscoveryPluginTestFramework(object):
    plugin = None
    path = None

    results_data_file = 'results.json'

    plugin_conf = {
        'Core': {
            'name': 'Test Plugin',
            'module': 'test_plugin'
        },
        'main': {
            'execute_frequency': '60',
            'enrichment_ttl': '300',
            'resource_filter': 'resource_class = "network"'
        }
    }

    def __init__(self, test_name):
        super(DiscoveryPluginTestFramework, self).__init__(test_name)

        self._path = self.path
        self._plugin_conf = self.plugin_conf

        self._expected_results = None
        self._results_data_file = 'data/' + self.results_data_file
        self._plugin_context = None
        self._panoptes_context = None

    def set_panoptes_context(self):
        panoptes_test_conf_file = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                               'config_files/test_panoptes_config.ini')

        self._panoptes_context = PanoptesContext(
                panoptes_test_conf_file
        )

    def set_plugin_context(self):
        self._plugin_context = create_autospec(
                PanoptesPluginContext, instance=True, spec_set=True,
                config=self._plugin_conf,
                logger=logging.getLogger(__name__)
        )

    def set_expected_results(self):
        expected_result_file = os.path.join(os.path.abspath(self.path), self._results_data_file)
        self._expected_results = json.load(open(expected_result_file))

    @patch('time.time', mock_time)
    @patch('yahoo_panoptes.framework.resources.time', mock_time)
    def setUp(self):
        self.set_panoptes_context()
        self.set_plugin_context()
        self.set_expected_results()


class SNMPPluginTestFramework(object):
    plugin_class = None
    path = None

    snmp_host = '127.0.0.1'
    snmp_port = 10161
    snmp_timeout = 10
    snmp_retries = 1
    snmp_max_repetitions = 10
    snmp_community = 'public'
    snmp_failure_timeout = 1

    resource_id = 'test_id'
    resource_endpoint = '127.0.0.1'
    resource_plugin = 'dummy'
    resource_site = 'test_site'
    resource_class = 'network'
    resource_subclass = 'test_subclass'
    resource_type = 'test_type'

    results_data_file = 'results.json'
    enrichment_data_file = 'enrichment_data'

    plugin_conf = {
        'Core': {
            'name': 'Test Plugin',
            'module': 'test_plugin'
        },
        'main': {
            'execute_frequency': '60',
            'enrichment_ttl': '300',
            'resource_filter': 'resource_class = "network"'
        }
    }

    def __init__(self, test_name):
        super(SNMPPluginTestFramework, self).__init__(test_name)

        self._path = self.path
        self._plugin_conf = self.plugin_conf

        self._snmp_host = self.snmp_host
        self._snmp_port = self.snmp_port
        self._snmp_timeout = self.snmp_timeout
        self._snmp_retries = self.snmp_retries
        self._snmp_max_repetitions = self.snmp_max_repetitions
        self._snmp_community = self.snmp_community
        self._snmp_failure_timeout = self.snmp_failure_timeout

        self._resource_id = self.resource_id
        self._resource_endpoint = self.resource_endpoint
        self._resource_plugin = self.resource_plugin
        self._resource_site = self.resource_site
        self._resource_class = self.resource_class
        self._resource_subclass = self.resource_subclass
        self._resource_type = self.resource_type

        self._panoptes_context = None
        self._panoptes_resource = None

        self._snmp_conf = None
        self._expected_results = None
        self._results_data_file = 'data/' + self.results_data_file
        self._enrichment_data_file = 'data/' + self.enrichment_data_file
        self._secret_store = None
        self._plugin_context = None
        self._plugin_context_bad = None
        self._snmp_conf_bad = None
        self._enrichment_kv = None
        self._enrichment_cache = None

    def set_panoptes_context(self):
        panoptes_test_conf_file = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                               'config_files/test_panoptes_config.ini')

        self._panoptes_context = PanoptesContext(
                panoptes_test_conf_file,
                key_value_store_class_list=[PanoptesEnrichmentCacheKeyValueStore]
        )

    def set_panoptes_resource(self):
        self._panoptes_resource = PanoptesResource(
                resource_site=self._resource_site,
                resource_class=self._resource_class,
                resource_subclass=self._resource_subclass,
                resource_type=self._resource_type,
                resource_id=self._resource_id,
                resource_endpoint=self._resource_endpoint,
                resource_plugin=self._resource_plugin
        )

        self._panoptes_resource.resource_metadata['model'] = 'model'

    def set_enrichment_cache(self):
        self._enrichment_kv = self._panoptes_context.get_kv_store(PanoptesEnrichmentCacheKeyValueStore)
        enrichment_data_file = os.path.join(os.path.abspath(self.path), self._enrichment_data_file)

        if self._plugin_conf.get('enrichment'):
            self._enrichment_cache = PanoptesEnrichmentCache(
                    self._panoptes_context,
                    self._plugin_conf,
                    self._panoptes_resource
            )

            try:
                with open(enrichment_data_file) as enrichment_data:
                    for line in enrichment_data:
                        key, value = line.strip().split('=>')
                        self._enrichment_kv.set(key, value)
            except Exception as e:
                raise IOError('Failed to load enrichment data file {}: {}'.format(enrichment_data_file, repr(e)))

    def set_snmp_conf(self):
        self._snmp_conf = self._panoptes_context.config_object.snmp_defaults

    def set_secret_store(self):
        self._secret_store = create_autospec(PanoptesSecretsStore, instance=True, spec_set=True)
        self._secret_store.get_by_site.return_value = self._snmp_community

    def set_plugin_context(self):
        self._plugin_context = create_autospec(
                PanoptesPluginWithEnrichmentContext, instance=True, spec_set=True,
                data=self._panoptes_resource,
                enrichment=self._enrichment_cache,
                config=self._plugin_conf,
                snmp=self._snmp_conf,
                secrets=self._secret_store,
                logger=logging.getLogger(__name__)
        )

    def set_snmp_conf_bad(self):
        self._snmp_conf_bad = copy.copy(self._snmp_conf)
        self._snmp_conf_bad['port'] += 1
        self._snmp_conf_bad['timeout'] = self._snmp_failure_timeout

    def set_plugin_context_bad(self):
        self._plugin_context_bad = create_autospec(
                PanoptesPluginWithEnrichmentContext, instance=True, spec_set=True,
                data=self._panoptes_resource,
                enrichment=self._enrichment_cache,
                config=self._plugin_conf,
                snmp=self._snmp_conf_bad,
                secrets=self._secret_store,
                logger=logging.getLogger(__name__)
        )

    def set_expected_results(self):
        expected_result_file = os.path.join(os.path.abspath(self.path), self._results_data_file)
        self._expected_results = json.load(open(expected_result_file))

    @patch('time.time', mock_time)
    @patch('yahoo_panoptes.framework.resources.time', mock_time)
    @patch('redis.StrictRedis', PanoptesMockRedis)
    def setUp(self):
        self.set_panoptes_context()
        self.set_panoptes_resource()
        self.set_enrichment_cache()
        self.set_snmp_conf()
        self.set_secret_store()
        self.set_plugin_context()
        self.set_snmp_conf_bad()
        self.set_plugin_context_bad()
        self.set_expected_results()


class SNMPEnrichmentPluginTestFramework(SNMPPluginTestFramework):
    @patch('time.time', mock_time)
    @patch('yahoo_panoptes.framework.resources.time', mock_time)
    def test_enrichment_plugin_results(self):
        """Test plugin result and validate results with input data/results.json"""
        plugin = self.plugin_class()
        result = plugin.run(self._plugin_context)

        self.assertIsNotNone(result)

        result = ordered(json.loads(result.json()))
        expected = ordered(self._expected_results)

        self.assertEqual(result, expected)

    def test_enrichment_plugin_timeout(self):
        """Test plugin raises error during SNMP timeout"""
        plugin = self.plugin_class()

        with self.assertRaises(Exception):
            plugin.run(self._plugin_context_bad)


class SNMPPollingPluginTestFramework(SNMPPluginTestFramework):
    uses_polling_status = True

    @staticmethod
    def _remove_timestamps(results):
        metrics_group_json_strings = list()

        for metrics_group in results.metrics_groups:

            metrics_group_dict = json.loads(metrics_group.json)

            resource_dict = metrics_group_dict['resource']
            resource_dict.pop('resource_creation_timestamp')

            for metric in metrics_group_dict['metrics']:
                metric.pop('metric_creation_timestamp')

            metrics_group_dict.pop('metrics_group_creation_timestamp')
            metrics_group_dict['resource'] = resource_dict

            metrics_group_json_strings.append(metrics_group_dict)

        return metrics_group_json_strings

    def set_expected_results(self):
        self._expected_results = list()
        expected_result_file = os.path.join(os.path.abspath(self.path), self._results_data_file)
        with open(expected_result_file) as results_file:
            expected_results = json.load(results_file)
            for result in expected_results:
                self._expected_results.append(result)

    def test_basic_operations(self):
        plugin = self.plugin_class()
        results = plugin.run(self._plugin_context)

        self.assertEqual(ordered(self._expected_results), ordered(self._remove_timestamps(results)))

    def test_invalid_resource_endpoint(self):
        self._resource_endpoint = '127.0.0.257'
        self._snmp_conf['timeout'] = self._snmp_failure_timeout
        self.set_panoptes_resource()
        self.set_plugin_context()

        plugin = self.plugin_class()
        with self.assertRaises(PanoptesPluginRuntimeError):
            plugin.run(self._plugin_context)

        self._resource_endpoint = '127.0.0.1'
        self._snmp_conf['timeout'] = self.snmp_timeout
        self.set_panoptes_resource()
        self.set_plugin_context()

    def test_inactive_port(self):
        """Tests a valid resource_endpoint with an inactive port"""
        plugin = self.plugin_class()
        results = plugin.run(self._plugin_context_bad)

        if self.uses_polling_status is True:
            self.assertEqual(len(results.metrics_groups), 1)
        else:
            self.assertEqual(len(results.metrics_groups), 0)

    def test_no_service_active(self):
        """Tests a valid resource_endpoint with no service active"""
        self._resource_endpoint = '127.0.0.2'
        self._snmp_conf['timeout'] = self._snmp_failure_timeout
        self.set_panoptes_resource()
        self.set_plugin_context()

        plugin = self.plugin_class()
        results = plugin.run(self._plugin_context)

        if self.uses_polling_status is True:
            self.assertEqual(len(results.metrics_groups), 1)
        else:
            self.assertEqual(len(results.metrics_groups), 0)

        self._resource_endpoint = '127.0.0.1'
        self._snmp_conf['timeout'] = self._snmp_timeout
        self.set_panoptes_resource()
        self.set_plugin_context()



"""
This module implements an abstract Message Producer based on Kafka Queues
"""
from kafka import KeyedProducer
from kafka.partitioner import Murmur2Partitioner

from ..validators import PanoptesValidators


class PanoptesMessageQueueProducer(object):
    """
    Creates a message producer client instance

    Args:
        panoptes_context (PanoptesContext): The PanoptesContext to use. The Kafka client associated with the context \
        would be used to create the message producer
        async(bool): Whether the created message producer should be asynchronous. Defaults to False. If you create an \
        asynchronous message producer, a background thread would be created in the process to handle the message sending
    """

    def __init__(self, panoptes_context, async=False):
        self._kafka_client = panoptes_context.kafka_client
        self._kafka_producer = KeyedProducer(self._kafka_client, async=async, partitioner=Murmur2Partitioner)

    def __del__(self):
        self.stop()

    def send_messages(self, topic, key, messages, partitioning_key=None):
        """
        Send messages to the specified topic

        This method tries to ensures that the topic exists before trying to send a message to it. If auto-creation of
        topics is enabled, then this should always succeed (barring Kafka/Zookeeper failures)

        Args:
            topic (str): The topic to which the message should be sent
            key (str): The key for the message
            messages (str): The message to send
            partitioning_key (str): If provided, then it would be used to select the partition for the message instead \
            of the message key

        Returns:
            None: Nothing. Passes through exceptions in case of failure

        """
        assert PanoptesValidators.valid_nonempty_string(topic), 'topic must be a non-empty string'
        assert PanoptesValidators.valid_nonempty_string(key), 'key must be a non-empty string'
        assert PanoptesValidators.valid_nonempty_string(messages), 'messages must be a non-empty string'

        self._kafka_client.ensure_topic_exists(topic)

        if partitioning_key:
            # We do this hack so that the partitioning key can be different from the message key
            partition = self._kafka_producer._next_partition(topic, partitioning_key)
            self._kafka_producer._send_messages(topic, partition, messages, key=key)
        else:
            # In this case, the message key is used as the partitioning key
            self._kafka_producer.send_messages(topic, key, messages)

    def stop(self):
        """
        Stop the message producer

        This method should always be called as part of cleanups so that any pending messages (for async message \
        producers) can be flushed and the Kafka connection closed cleanly

        Returns:
            None: Nothing. Passes through exceptions in case of failure

        """
        if not self._kafka_producer.stopped:
            self._kafka_producer.stop()



"""
SNMP V2 mib class
"""
from .base import oid, Mib


class MibSNMPV2(Mib):
    """
    Mib class with snmp v2 oids
    """
    sysDescr = oid('.1.3.6.1.2.1.1.1')
    interfaces = oid('.1.3.6.1.2.1.2')



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module declares an abstract base class for Panoptes Enrichment Plugins - and related helpers

A Enrichment Plugin's role is to return a collection of enrichment info that have been collected by the system
"""
import abc

import six

from ..framework.plugins.context import PanoptesPluginContext
from ..polling.polling_plugin import PanoptesPollingPluginError,\
    PanoptesPollingPluginConfigurationError, PanoptesPollingPlugin, PanoptesPollingPluginInfo


class PanoptesEnrichmentPluginError(PanoptesPollingPluginError):
    """
    The class for Enrichment Plugin runtime errors
    """
    pass


class PanoptesEnrichmentPluginConfigurationError(PanoptesPollingPluginConfigurationError):
    """
    The class for Enrichment Plugin configuration errors
    """
    pass


class PanoptesEnrichmentPluginInfo(PanoptesPollingPluginInfo):
    """
    This class extends the PanoptesPollingPluginInfo to contain the resource_filter associated with an Enrichment
    Plugin

    Args:
        plugin_name(str): The name of the plugin
        plugin_path(str): The 'path' of the plugin - either the absolute path of the directory that contains the plugin\
        module or the path of the module file without the '.py' extension
    """
    pass


@six.add_metaclass(abc.ABCMeta)
class PanoptesEnrichmentPlugin(PanoptesPollingPlugin):
    """
    The base class for all Panoptes Enrichment plugins

    A Enrichment Plugin's role is to return a collection of enrichment info associated with a resource
    """

    @abc.abstractmethod
    def run(self, context):
        """
        The entry point for Enrichment plugins

        Args:
            context (PanoptesPluginContext): The context of the plugin contains the plugins configuration, the logger \
            it should use and a client for the system wide plugins KV store

        Returns:
            object: Enrichment data PanoptesEnrichmentGroupSet/PanoptesEnrichmentMultiGroupSet
        collected by the plugin

       Raises:
            PanoptesPluginConfigurationError: If the plugin cannot proceed with the configuration provided to it \
            through the context, it should raise a PanoptesPluginConfigurationError

        """
        pass



import time
from importlib import import_module

from ..plugins.panoptes_base_plugin import PanoptesBasePlugin, PanoptesPluginConfigurationError, \
    PanoptesPluginRuntimeError
from ..utilities.snmp.connection import PanoptesSNMPPluginConfiguration, PanoptesSNMPConnectionFactory


class PanoptesSNMPBasePlugin(PanoptesBasePlugin):
    def __init__(self):
        super(PanoptesSNMPBasePlugin, self).__init__()
        self._plugin_context = None
        self._plugin_config = None
        self._logger = None
        self._snmp_configuration = None
        self._execute_frequency = None
        self._resource = None
        self._enrichment = None
        self._host = None

    @property
    def plugin_context(self):
        return self._plugin_context

    @property
    def plugin_config(self):
        return self._plugin_config

    @property
    def logger(self):
        return self._logger

    @property
    def resource(self):
        return self._resource

    @property
    def enrichment(self):
        return self._enrichment

    @property
    def execute_frequency(self):
        return self._execute_frequency

    @property
    def host(self):
        return self._host

    @property
    def snmp_configuration(self):
        return self._snmp_configuration

    @property
    def snmp_connection(self):
        return self._snmp_connection

    def _get_snmp_connection(self):
        snmp_connection_class = getattr(import_module(self.snmp_configuration.connection_factory_module),
                                        self.snmp_configuration.connection_factory_class)

        assert issubclass(snmp_connection_class,
                          PanoptesSNMPConnectionFactory), 'SNMP connection class must be a subclass of ' \
                                                          'PanoptesSNMPConnectionFactory'

        self._snmp_connection = snmp_connection_class.get_snmp_connection(plugin_context=self._plugin_context,
                                                                          snmp_configuration=self._snmp_configuration)

    def get_results(self):
        return

    def run(self, context):
        self._plugin_context = context
        self._plugin_config = context.config
        self._logger = context.logger
        self._execute_frequency = int(self._plugin_config['main']['execute_frequency'])
        self._resource = context.data
        self._enrichment = context.enrichment
        self._host = self._resource.resource_endpoint

        try:
            self._snmp_configuration = PanoptesSNMPPluginConfiguration(self._plugin_context)
        except Exception as e:
            raise PanoptesPluginConfigurationError('Error parsing SNMP configuration: {}'.format(repr(e)))

        try:
            self._get_snmp_connection()
        except Exception as e:
            raise PanoptesPluginRuntimeError('Error creating SNMP connection: {}'.format(repr(e)))

        start_time = time.time()

        self.logger.info('Going to poll device "{}:{}" for metrics'.format(self._host, self.snmp_configuration.port))

        results = self.get_results()

        end_time = time.time()

        if results:
            self.logger.info(
                    'Done polling metrics for device "{}" in {:.2f} seconds, {} metrics groups'.format(
                            self._host, end_time - start_time, len(results)))
        else:
            self.logger.warn('Error polling metrics for device {}'.format(self._host))

        return results


class PanoptesSNMPBaseEnrichmentPlugin(PanoptesSNMPBasePlugin):
    def __init__(self):
        super(PanoptesSNMPBaseEnrichmentPlugin, self).__init__()
        self._enrichment_ttl = None

    @property
    def enrichment_ttl(self):
        return self._enrichment_ttl

    def run(self, context):
        self._enrichment_ttl = int(context.config['main']['enrichment_ttl'])
        return super(PanoptesSNMPBaseEnrichmentPlugin, self).run(context)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module implements Neighbor Lookup EnrichmentGroup defined with schema validator
"""

from ...framework.enrichment import PanoptesEnrichmentSchemaValidator, PanoptesEnrichmentGroup

INTERFACE_LOOKUP_SCHEMA_NAMESPACE = 'interface_lookup'
DEVICE_LOOKUP_SCHEMA_NAMESPACE = 'device_lookup'


class PanoptesInterfaceLookupEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'resource_id': {'type': 'string', 'required': True},
                'interface_description': {'type': 'string', 'required': True},
                'interface_index': {'type': 'string', 'required': True},
                'port_id_map': {'type': 'dict', 'required': True}
            }
        }
    }


class PanoptesDeviceLookupEnrichmentSchemaValidator(PanoptesEnrichmentSchemaValidator):
    schema = {
        'enrichment_label': {
            'type': 'dict',
            'schema': {
                'device_info_ipv4': {'type': 'dict', 'required': True},
                'device_info_ipv6': {'type': 'dict', 'required': True},
                'source_resource_id': {'type': 'string', 'required': True}
            }
        }
    }


class PanoptesInterfaceLookupEnrichmentGroup(PanoptesEnrichmentGroup):
    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesInterfaceLookupEnrichmentGroup, self).__init__(
            namespace=INTERFACE_LOOKUP_SCHEMA_NAMESPACE,
            schema_validator=PanoptesInterfaceLookupEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)


class PanoptesDeviceLookupEnrichmentGroup(PanoptesEnrichmentGroup):
    def __init__(self, enrichment_ttl, execute_frequency):
        super(PanoptesDeviceLookupEnrichmentGroup, self).__init__(
            namespace=DEVICE_LOOKUP_SCHEMA_NAMESPACE,
            schema_validator=PanoptesDeviceLookupEnrichmentSchemaValidator(),
            enrichment_ttl=enrichment_ttl,
            execute_frequency=execute_frequency)



"""
This module supplies an ICMP ping utility. Only supports Linux systems.
"""

import json
import re
import subprocess

from ..exceptions import PanoptesBaseException
from ..validators import PanoptesValidators

_VALID_PING_STATS = re.compile(r'.*ping statistics ---\n([0-9]*) packets transmitted, ([0-9]*) received, '
                               r'(\d*\.?\d*)% packet loss, time ([0-9]*)ms\nrtt min/avg/max/mdev = '
                               r'(\d*\.\d+)/(\d*\.\d+)/(\d*\.\d+)/(\d*\.\d+) ms')
"""
    regular expression object: Precompile regex pattern for collecting ping statistics
"""


class PanoptesPingException(PanoptesBaseException):
    """
        The class for all ping exceptions in the Panoptes system
    """
    pass


class PanoptesPingTimeoutException(PanoptesPingException):
    """
        The class for all ping timeout exceptions in the Panoptes system
    """
    pass


class PanoptesPing(object):
    """
        The class for pinging a device and returning its status

        Args:
            count (int): number of ping attempts to be made
            hostname (str): name of the host to ping
            timeout (int): time to wait before cancelling ping
    """

    def __init__(self, count=10, timeout=10, hostname="localhost"):
        assert PanoptesValidators.valid_nonzero_integer(count), 'count must be integer > 0'
        assert PanoptesValidators.valid_nonempty_string(hostname), 'hostname must be nonempty string'
        assert PanoptesValidators.valid_nonzero_integer(timeout), 'timeout must be integer > 0'

        self._response = dict()  # dictionary containing ping statistics
        self._response['packets_transmitted'] = None
        self._response['packets_received'] = None
        self._response['packet_loss_pct'] = None
        self._response['execution_time'] = None
        self._response['round_trip_min'] = None
        self._response['round_trip_avg'] = None
        self._response['round_trip_max'] = None
        self._response['round_trip_stddev'] = None

        try:
            resp = subprocess.check_output(
                ['/bin/ping', '-c', str(count), '-w', str(timeout), hostname],
                stderr=subprocess.STDOUT,
                universal_newlines=True  # return string not bytes
            )
            self._get_ping_stats(resp)
        except subprocess.CalledProcessError as e:
            self._get_ping_stats(e.output)
            if self._response['packets_transmitted'] is not None:
                raise PanoptesPingTimeoutException("Ping timed out with response: " + str(self._response))
            raise PanoptesPingException(e.output)
        except Exception as e:
            raise PanoptesPingException(e.message)

    def _get_ping_stats(self, resp):
        m = _VALID_PING_STATS.search(resp)
        if m:
            self._response['packets_transmitted'] = int(m.group(1))
            self._response['packets_received'] = int(m.group(2))
            self._response['packet_loss_pct'] = float(m.group(3))
            self._response['execution_time'] = round(float(m.group(4)) / 1000, 2)  # in seconds
            self._response['round_trip_min'] = float(m.group(5))
            self._response['round_trip_avg'] = float(m.group(6))
            self._response['round_trip_max'] = float(m.group(7))
            self._response['round_trip_stddev'] = float(m.group(8))

    @property
    def packets_transmitted(self):
        """whole number count of packets transmitted; should equal 'count'"""
        return self._response['packets_transmitted']

    @property
    def packets_received(self):
        """whole number count of packets received"""
        return self._response['packets_received']

    @property
    def packet_loss_pct(self):
        """percentage of packets lost"""
        return self._response['packet_loss_pct']

    @property
    def execution_time(self):
        """duration in seconds of time 'ping' takes to execute"""
        return self._response['execution_time']

    @property
    def round_trip_min(self):
        """minimum round-trip time"""
        return self._response['round_trip_min']

    @property
    def round_trip_avg(self):
        """average round-trip time"""
        return self._response['round_trip_avg']

    @property
    def round_trip_max(self):
        """maximum round-trip time"""
        return self._response['round_trip_max']

    @property
    def round_trip_stddev(self):
        """standard deviation of all round-trip times"""
        return self._response['round_trip_stddev']

    @property
    def response(self):
        """returns JSON of all ping statistics"""
        return json.dumps(self._response)



from yahoo_panoptes.polling.polling_plugin import PanoptesPollingPlugin


class PanoptesTestPollingPlugin(PanoptesPollingPlugin):
    def run(self, context):
        super(PanoptesTestPollingPlugin, self).run(context)



import json
import sys
import time

from json_schema_validator.errors import ValidationError
from json_schema_validator.schema import Schema
from json_schema_validator.validator import Validator
from kafka import KafkaConsumer
from kafka.common import OffsetAndMetadata

from ..context import PanoptesContext
from ..validators import PanoptesValidators


class PanoptesConsumerTypes(object):
    """
    This class contains the supported consumer types as attributes
    """
    METRICS, RESOURCES, PROCESSED = range(3)


CONSUMER_TYPE_NAMES = dict((getattr(PanoptesConsumerTypes, n), n.lower())
                           for n in dir(PanoptesConsumerTypes) if '_' not in n)


def make_topic_names_for_all_sites(panoptes_context, topic_suffix):
    delimiter = panoptes_context.config_dict['kafka']['topic_name_delimiter']
    return [delimiter.join([site, topic_suffix]) for site in panoptes_context.config_object.sites]


def get_consumer_type_from_name(name):
    return getattr(PanoptesConsumerTypes, name.upper())


class PanoptesConsumerRecordValidator(object):
    """
    This class implements validators for various consumer types
    """

    _metrics_schema = Schema(
            {
                "$schema": "http://json-schema.org/draft-04/schema#",
                "type": "object",
                "properties": {
                    "metrics": {
                        "type": "array",
                        "items": [{
                            "type": "object",
                            "properties": {
                                "metric_name": {"type": "string"},
                                "metric_value": {"type": "number"},
                                "metric_type": {"type": "string", "enum": ["gauge", "counter"]},
                                "metric_creation_timestamp": {"type": "number"}
                            },
                            "required": ["metric_name", "metric_value", "metric_type", "metric_creation_timestamp"]
                        }],
                        "minItems": 1
                    },
                    "dimensions": {
                        "type": "array",
                        "items": [{
                            "type": "object",
                            "properties": {
                                "dimension_name": {"type": "string"},
                                "dimension_value": {"type": "string"}
                            },
                            "required": ["dimension_name", "dimension_value"]
                        }]
                    },
                    "resource": {
                        "type": "object",
                        "properties": {
                            "resource_site": {"type": "string"},
                            "resource_class": {"type": "string"},
                            "resource_subclass": {"type": "string"},
                            "resource_type": {"type": "string"},
                            "resource_id": {"type": "string"}
                        },
                        "required": ["resource_site", "resource_class", "resource_subclass", "resource_type",
                                     "resource_id"]
                    },
                    "metrics_group_type": {"type": "string"},
                    "metrics_group_interval": {"type": "number"},
                    "metrics_group_creation_timestamp": {"type": "number"},
                    "metrics_group_schema_version": {"type": "string", "enum": ["0.2"]}
                },
                "required": ["metrics", "dimensions", "resource", "metrics_group_type", "metrics_group_interval",
                             "metrics_group_creation_timestamp", "metrics_group_schema_version"]
            }
    )

    _resource_schema = Schema(
            {
                "$schema": "http://json-schema.org/draft-04/schema#",
                "type": "object",
                "properties": {
                    "resources": {
                        "type": "array",
                        "items": [{
                            "type": "object",
                            "properties": {
                                "resource_site": {"type": "string"},
                                "resource_class": {"type": "string"},
                                "resource_subclass": {"type": "string"},
                                "resource_type": {"type": "string"},
                                "resource_id": {"type": "string"},
                                "resource_endpoint": {"type": "string"},
                                "resource_creation_timestamp": {"type": "number"},
                                "resource_metadata": {
                                    "type": "object",
                                    "patternProperties": {
                                        "^[a-zA-Z0-9][a-zA-Z0-9]*$": {"type": "string"}
                                    }
                                }
                            },
                            "required": ["resource_site", "resource_class", "resource_subclass", "resource_type",
                                         "resource_id", "resource_endpoint", "resource_creation_timestamp"]
                        }],
                        "minItems": 1
                    },
                    "resource_set_creation_timestamp": {"type": "number"},
                    "resource_set_schema_version": {"type": "string", "enum": ["0.1"]}
                },
                "required": ["resources", "resource_set_creation_timestamp", "resource_set_schema_version"]
            }
    )

    @staticmethod
    def validate_metrics(consumer_record):
        """
        This method validates that the passed consumer record  is a valid metrics group record

        Args:
            consumer_record (str): A valid JSON string representation of the consumer record

        Returns:
            bool: True if validation passes, False otherwise
        """
        try:
            return Validator.validate(PanoptesConsumerRecordValidator._metrics_schema, consumer_record)
        except ValidationError:
            return False

    @staticmethod
    def validate_resources(consumer_record):
        """
        This method validates that the passed consumer record  is a valid resources record

        Args:
            consumer_record (str): A valid JSON string representation of the consumer record

        Returns:
            bool: True if validation passes, False otherwise
        """
        try:
            return Validator.validate(PanoptesConsumerRecordValidator._resource_schema, consumer_record)
        except ValidationError:
            return False

    @staticmethod
    def validate(consumer_type, consumer_record):
        """
        This method is the 'entry point' for this class and routes the validation to type specific methods

        Args:
            consumer_type (int): A valid consumer type
            consumer_record (str): A valid JSON string representation of the consumer record
        """
        if consumer_type == PanoptesConsumerTypes.METRICS:
            return PanoptesConsumerRecordValidator.validate_metrics(consumer_record)
        elif consumer_type == PanoptesConsumerTypes.RESOURCES:
            return PanoptesConsumerRecordValidator.validate_resources(consumer_record)
        elif consumer_type == PanoptesConsumerTypes.PROCESSED:
            return PanoptesConsumerRecordValidator.validate_metrics(consumer_record)
        else:
            return False


class PanoptesConsumer(object):
    def __init__(self, panoptes_context, consumer_type, topics, client_id, group, keys, poll_timeout, callback,
                 validate=False, session_timeout=60, max_poll_records=500, max_partition_fetch_bytes=1048576):
        """
        This class implements a helper/wrapper for writing Panoptes Consumers

        The consumer object produced by this class joins the group provided in the arguments, subscribes to the relevant
        topics (which are combination of the site name and consumer type), filters records by the provided keys, checks
        validity of the records by converting them to JSON and validating against consumer type specific schemas and
        then calls the callback function provided for each record.

        The consumer object also takes care of advancing the committed marker for each topic and partition - if the
        callback fails or returns false, the marker is not advanced.

        Args:
            panoptes_context (PanoptesContext): The PanoptesContext to be used by the consumer
            consumer_type (int): A valid consumer type
            topics (list): A valid consumer type
            client_id (str): A non-empty string that uniquely identifies the consumer instance - this is used for \
            logging by and group administration by Kafka
            group (str): The Kafka group this consumer should be a part of
            keys (list, None): A comma separated list of keys the consumer should filter against
            poll_timeout (int): A non-negative integer which is the interval (in seconds) the consumer should sleep if \
            no records are available on the Kafka bus
            callback (callable): A callable to which each processed, validated consumer records (deserialized JSON \
            object) should be passed. The callable should return false if it cannot process the record due to
            temporary \
            issues - it would be redelivered to the callback in that case
            validate (bool): Whether each incoming record should be validated. Defaults to False
            session_timeout (int): A non-negative integer which is the interval (in seconds) after which the Kafka \
            Group Management system should consider the client disconnected
            max_poll_records (int): The maximum number of records to fetch in one poll cycle

        Returns:
            None
        """
        assert isinstance(panoptes_context, PanoptesContext), 'panoptes_context must be an instance of PanoptesContext'
        assert consumer_type in CONSUMER_TYPE_NAMES, 'consumer_type must be an valid attribute from ' \
                                                     'PanoptesConsumerTypes'
        assert PanoptesValidators.valid_nonempty_iterable_of_strings(topics), ''
        assert isinstance(client_id, str), 'client_id must be a non-empty string'
        assert keys is None or PanoptesValidators.valid_nonempty_iterable_of_strings(keys), \
            'keys must be None or a list of non-empty strings'
        assert PanoptesValidators.valid_positive_integer(poll_timeout), 'poll_timeout must be an integer'
        assert hasattr(callback, '__call__'), 'callback must be a callable'
        assert isinstance(validate, bool), 'validate must be a boolean'
        assert PanoptesValidators.valid_nonzero_integer(session_timeout), 'session_timeout must be an integer '\
                                                                          'greater than zero'
        assert PanoptesValidators.valid_nonzero_integer(max_poll_records), 'max_poll_records must be an integer '\
                                                                           'greater than zero'

        self._panoptes_context = panoptes_context
        self._consumer_type = consumer_type
        self._topics = topics
        self._client_id = client_id
        self._keys = keys
        self._group = group
        self._poll_timeout = poll_timeout * 1000
        self._session_timeout = session_timeout * 1000
        self._request_timeout = self._session_timeout * 3
        self._heartbeat_interval = self._session_timeout / 3
        self._max_poll_records = max_poll_records
        self._max_partition_fetch_bytes = max_partition_fetch_bytes
        self._callback = callback
        self._validate = validate

        self._consumer = None
        self._last_polled = 0
        self._asked_to_stop = False

    @property
    def panoptes_context(self):
        """
        The Panoptes Context associated with this consumer

        Returns:
            PanoptesContext:  The Panoptes Context associated with this consumer
        """
        return self._panoptes_context

    @property
    def client_id(self):
        """
        The Client Id associated with this consumer

        Returns:
            str: The Client Id associated with this consumer
        """
        return self._client_id

    @property
    def group(self):
        """
        The Kafka group associated with this consumer

        Returns:
            str: The Kafka group associated with this consumer
        """
        return self._group

    @property
    def poll_timeout(self):
        """
        The Kafka poll timeout (in seconds) set for this consumer

        Returns:
            int: The Kafka poll timeout (in seconds) set for this consumer
        """
        return self._poll_timeout

    @property
    def consumer_type(self):
        """
        The type of consumer

        Returns:
            int: The type of consumer this object is
        """
        return self._consumer_type

    @property
    def keys(self):
        """
        The comma separated list of keys associated with this consumer

        Returns:
            str: he comma separated list of keys associated with this consumer
        """
        return self._keys

    def start_consumer(self):
        """
        This method is the workhorse of this class - it starts the Kafka consumer and calls the callback function for
        each valid record
        """
        logger = self.panoptes_context.logger
        config = self.panoptes_context.config_object
        last_batch_size = 0

        logger.info('Trying to start Kafka Consumer with brokers: "%s", topics: "%s", group: "%s"' % (
            config.kafka_brokers, self._topics, self.group))

        try:
            consumer = KafkaConsumer(bootstrap_servers=config.kafka_brokers,
                                     client_id=self.client_id,
                                     group_id=self.group,
                                     enable_auto_commit=False,
                                     session_timeout_ms=self._session_timeout,
                                     request_timeout_ms=self._request_timeout,
                                     heartbeat_interval_ms=self._heartbeat_interval,
                                     max_poll_records=self._max_poll_records,
                                     max_partition_fetch_bytes=self._max_partition_fetch_bytes)
            consumer.subscribe(topics=self._topics)
            logger.info('Consumer subscribed to: %s' % consumer.subscription())
            self._consumer = consumer
        except Exception as e:
            sys.exit('Error trying to start Kafka consumer: %s' % str(e))

        while not self._asked_to_stop:
            poll_age = (time.time() - self._last_polled) * 1000
            if (poll_age > self._session_timeout) and (last_batch_size > 0):
                logger.warn('Poll cycle took %.2f ms for %d records, '
                            'which is greater than the session timeout of %d ms' %
                            (poll_age, last_batch_size, self._session_timeout))

            try:
                topic_partitions = consumer.poll(timeout_ms=self._poll_timeout)
                self._last_polled = time.time()
                logger.debug('Poll returned with %d topic partitions' % len(topic_partitions))
            except Exception as e:
                logger.error('Error while polling: %s' % str(e))
                continue

            last_batch_size = 0
            for topic_partition in topic_partitions.keys():
                consumer_records = topic_partitions[topic_partition]
                last_batch_size += len(consumer_records)

                logger.debug('Processing topic partition: %s, consumer records: %d, committed: %s' % (
                    str(topic_partition), len(consumer_records), consumer.committed(topic_partition)))
                logger.debug('Consumed offsets: %s' % consumer._subscription.all_consumed_offsets())

                callback_succeeded = True
                consumer_records_skipped = 0
                consumer_records_validation_failed = 0

                for consumer_record in consumer_records:
                    logger.debug('Processing consumer record with key: "%s" and value: "%s"' % (
                        consumer_record.key, consumer_record.value))

                    if self.keys and consumer_record.key not in self.keys:
                        logger.debug(
                                'Consumer record key "%s" does not match any of the provided keys, skipping' %
                                consumer_record.key)
                        consumer_records_skipped += 1
                        continue

                    try:
                        consumer_record_object = json.loads(consumer_record.value)
                    except Exception as e:
                        logger.warn(
                                'Could not convert consumer record "%s" to JSON, skipping: %s' % (
                                    consumer_record.value, str(e)))
                        consumer_records_validation_failed += 1
                        continue

                    if self._validate:
                        if not PanoptesConsumerRecordValidator.validate(
                                self.consumer_type, consumer_record_object):
                            logger.debug('Consumer record failed validation, skipping')
                            consumer_records_validation_failed += 1
                            continue

                    try:
                        callback_succeeded = self._callback(consumer_record.key, consumer_record_object)
                        # If the callback fails even for one consumer record, we want to fail (not update the committed)
                        # offset for the entire the batch, so exit
                        if not callback_succeeded:
                            logger.error('Callback function returned false')
                            break
                    except:
                        logger.exception('Error trying to execute callback function')
                        break

                # Update the committed offset if the callback function succeeds for *all* consumer records in this topic
                # partition
                if callback_succeeded:
                    try:
                        position = consumer.position(topic_partition)
                    except Exception as e:
                        logger.error(
                                'Error trying to fetch position for topic partition "%s": %s' % (
                                    topic_partition, str(e)))
                    else:
                        offset = {topic_partition: OffsetAndMetadata(offset=position, metadata='')}
                        logger.debug('Going to commit offset %s' % str(offset))
                        try:
                            consumer.commit(offset)
                        except Exception as e:
                            logger.error('Error trying to commit offset "%s": %s' % (offset, str(e)))

                if consumer_records_skipped or consumer_records_validation_failed:
                    logger.debug(
                            'Skipped %d consumer records due to non-matching keys and %d consumer records due to '
                            'validation failures for topic partition: %s' % (
                                consumer_records_skipped, consumer_records_validation_failed, topic_partition))

    def stop_consumer(self):
        """
        Stops the consumer gracefully
        """
        self._asked_to_stop = True
        if self._consumer:
            self._consumer.unsubscribe()
            self._consumer.close()


class PanoptesResourcesConsumer(PanoptesConsumer):
    def __init__(self, panoptes_context, client_id, group, keys, poll_timeout, callback,
                 validate=False, session_timeout=30, max_poll_records=500, max_partition_fetch_bytes=1048576):
        """
        This class implements a helper/wrapper for writing Panoptes Consumers

        The consumer object produced by this class joins the group provided in the arguments, subscribes to the
        relevant
        topics (which are combination of the site name and consumer type), filters records by the provided keys,
        checks
        validity of the records by converting them to JSON and validating against consumer type specific schemas and
        then calls the callback function provided for each record.

        The consumer object also takes care of advancing the committed marker for each topic and partition - if the
        callback fails or returns false, the marker is not advanced.

        Args:
            panoptes_context (PanoptesContext): The PanoptesContext to be used by the consumer
            client_id (str): A non-empty string that uniquely identifies the consumer instance - this is used for \
            logging by and group administration by Kafka
            group (str): The Kafka group this consumer should be a part of
            keys (list, None): A comma separated list of keys the consumer should filter against
            poll_timeout (int): A non-negative integer which is the interval (in seconds) the consumer should
            sleep if \
            no records are available on the Kafka bus
            callback (callable): A callable to which each processed, validated consumer records (deserialized JSON \
            object) should be passed. The callable should return false if it cannot process the record due to
            temporary \
            issues - it would be redelivered to the callback in that case
            validate (bool): Whether each incoming record should be validated. Defaults to False
            session_timeout (int): A non-negative integer which is the interval (in seconds) after which the Kafka \
            Group Management system should consider the client disconnected
            max_poll_records (int): The maximum number of records to fetch in one poll cycle

        Returns:
            None
        """
        topics = make_topic_names_for_all_sites(panoptes_context,
                                                CONSUMER_TYPE_NAMES[PanoptesConsumerTypes.RESOURCES])

        super(PanoptesResourcesConsumer, self).__init__(panoptes_context=panoptes_context,
                                                        consumer_type=PanoptesConsumerTypes.RESOURCES,
                                                        topics=topics,
                                                        client_id=client_id,
                                                        group=group,
                                                        keys=keys,
                                                        poll_timeout=poll_timeout,
                                                        callback=callback,
                                                        validate=validate,
                                                        session_timeout=session_timeout,
                                                        max_poll_records=max_poll_records,
                                                        max_partition_fetch_bytes=max_partition_fetch_bytes)



"""
Copyright 2018, Oath Inc.
Licensed under the terms of the Apache 2.0 license. See LICENSE file in project root for terms.

This module defines resources and their related abstractions
"""
import hashlib
import json
import re
import sqlite3
import threading
from collections import OrderedDict
from time import time

from cached_property import threaded_cached_property
from pyparsing import CaselessKeyword, Group, Word, oneOf, delimitedList, ZeroOrMore, Forward, \
    ParseException, Optional, QuotedString, CaselessLiteral, Literal, alphanums, ParseResults, upcaseTokens, \
    downcaseTokens
from six import string_types, integer_types

from . import const
from .context import PanoptesContext
from .exceptions import PanoptesBaseException
from .utilities.key_value_store import PanoptesKeyValueStore
from .validators import PanoptesValidators


class PanoptesResourceError(PanoptesBaseException):
    """
    The base class for all Panoptes Resources errors
    """
    pass


class PanoptesResourceCacheException(PanoptesResourceError):
    """
    The base class for PanoptesResourceCache errors
    """
    pass


class PanoptesResource(object):
    """
    Representation of the a device/endpoint that should be monitored

    A resource in Panoptes is uniquely defined by it's class, subclass, type and id. Class, subclass and type jointly \
    act as the namespace of the resource - the id must be unique within this namespace

    Args:
        resource_site (str): The name of the site to which the resource belongs (e.g. colo name)
        resource_class (str): The class of the resource (e.g. 'network')
        resource_subclass (str): The subclass of the resource (e.g. 'load-balancer')
        resource_type (str): The type of the resource (e.g. 'a10')
        resource_id (str): A unique identifier for the resource
        resource_endpoint (str): The endpoint that should be polled by the monitoring system for the resource
        resource_creation_timestamp (float): A  UTC Unix Epoch timestamp of when the Resource was created. If this is
        not provided, then the UTC timestamp of the current time is automatically added
        resource_plugin (str): The name of the discovery plugin that found this resource
        resource_ttl(int): Resource cache ttl in seconds

    Notes:
        - All the parameters (except resource_endpoint and resource_creation_timestamp) are
        required and cannot be null or empty
        - The class, subclass, type and id are arbitrary strings and not validated
        - Please do not use randomly generated UUIDs/GUIDs for the id of resource - the reason is that you create
        another resource with exactly the same data everywhere except the id, Panoptes  will treat these as
        two separate resources throughout the system. If you plan to use UUIDs/GUIDs, please make sure that they come
        from a central source of truth/registry and that the same resource is assigned the same UUID/GUID every time
        the resource object is created
    """

    _metadata_key = re.compile(r"^[^\d\W]\w*\Z")

    def __init__(self, resource_site, resource_class, resource_subclass, resource_type,
                 resource_id, resource_endpoint, resource_creation_timestamp=None,
                 resource_plugin=None, resource_ttl=const.RESOURCE_MANAGER_RESOURCE_EXPIRE):
        assert resource_site and isinstance(resource_site,
                                            string_types), 'resource_site must be a non-empty str or unicode'
        assert resource_class and isinstance(resource_class,
                                             string_types), 'resource_class must be a non-empty str or unicode'
        assert resource_subclass and isinstance(resource_subclass,
                                                string_types), 'resource_subclass must be a non-empty str or unicode'
        assert resource_type and isinstance(resource_type,
                                            string_types), 'resource_type must be a non-empty str or unicode'
        assert resource_id and isinstance(resource_id,
                                          string_types), 'resource_id must be a non-empty str or unicode'
        assert resource_endpoint and isinstance(resource_endpoint,
                                                string_types), 'resource_endpoint must be a non-empty str or unicode'
        assert resource_plugin is None or isinstance(resource_plugin,
                                                     string_types), 'resource_plugin must be None or a non-empty str ' \
                                                                    'or unicode'
        assert resource_ttl > 0 and isinstance(resource_ttl, integer_types), \
            'resource_ttl must be an integer greater than 0'

        self.__data = OrderedDict()
        self.__data['resource_site'] = str(resource_site)
        self.__data['resource_class'] = str(resource_class)
        self.__data['resource_subclass'] = str(resource_subclass)
        self.__data['resource_type'] = str(resource_type)
        self.__data['resource_id'] = str(resource_id)
        self.__data['resource_endpoint'] = str(resource_endpoint)
        self.__data['resource_metadata'] = OrderedDict()
        if not resource_creation_timestamp:
            self.__data['resource_creation_timestamp'] = time()
        else:
            self.__data['resource_creation_timestamp'] = resource_creation_timestamp
        self.__data['resource_plugin'] = resource_plugin
        self.__data['resource_metadata']['_resource_ttl'] = str(resource_ttl)

    def add_metadata(self, key, value):
        """
        Add a metadata key/value pair

        Args:
            key(str): The syntax for a key is the same as for valid Python identifier: (letter|"_") (letter | digit "_")
            value(str): A pipe (|) is a special reserved character and cannot be present in a value

        Returns:
            None

        Raises:
            ValueError: If the key or value don't match the acceptable string patterns, a ValueError is raised

        Examples:
            add_metadata('os_name', 'Advanced Core OS')
            add_metadata('os_version', '2.6.1-GR1-P16')
        """
        assert PanoptesValidators.valid_nonempty_string(key), 'key must be a non-empty str or unicode'
        assert PanoptesValidators.valid_nonempty_string(value), 'value must be a non-empty str or unicode'

        if not self.__class__._metadata_key.match(key):
            raise ValueError('metadata key "%s" has to match pattern: (letter|"_") (letter | digit | "_")*' % key)

        if '|' in value:
            raise ValueError('metadata value "%s" cannot contain |' % value)

        self.__data['resource_metadata'][key] = value

    @property
    def resource_site(self):
        """
        The site name to which the resource belongs

        Returns:
            str

        """
        return self.__data['resource_site']

    @property
    def resource_class(self):
        """
        The class of the resource

        Returns:
            str

        """
        return self.__data['resource_class']

    @property
    def resource_subclass(self):
        """
        The subclass of the resources

        Returns:
            str

        """
        return self.__data['resource_subclass']

    @property
    def resource_type(self):
        """
        The type of the resource

        Returns:
            str

        """
        return self.__data['resource_type']

    @property
    def resource_id(self):
        """
        The id of the resource

        Returns:
            str

        """
        return self.__data['resource_id']

    @property
    def resource_endpoint(self):
        """
        The endpoint to be monitored for the resource

        Returns:
            str

        """
        return self.__data['resource_endpoint']

    @property
    def resource_ttl(self):
        """
        The cache ttl seconds of the resource

        Returns:
            str

        """
        return self.__data['resource_metadata']['_resource_ttl']

    @property
    def resource_creation_timestamp(self):
        return self.__data['resource_creation_timestamp']

    @property
    def resource_plugin(self):
        return self.__data['resource_plugin']

    @threaded_cached_property
    def serialization_key(self):
        key = const.KV_STORE_DELIMITER.join(
                ['plugin', self.resource_plugin, 'site', self.resource_site, 'class', self.resource_class,
                 'subclass', self.resource_subclass, 'type', self.resource_type,
                 'id', self.resource_id, 'endpoint', self.resource_endpoint])

        return key

    @property
    def resource_metadata(self):
        """
        The metadata associated with the resource

        Returns:
            dict

        """
        return self.__data['resource_metadata']

    @property
    def json(self):
        """
        The JSON representation of the resource

        Returns:
            str: The JSON representation of the resource
        """
        return json.dumps(self.__data)

    @property
    def raw(self):
        return self.__data

    def __repr__(self):
        return str(self.serialization_key)

    def __hash__(self):
        return int(hashlib.md5(self.serialization_key).hexdigest(), 16)

    def __eq__(self, other):
        if not isinstance(other, PanoptesResource):
            return False
        return (
            self.resource_site == other.resource_site and
            self.resource_class == other.resource_class and
            self.resource_subclass == other.resource_subclass and
            self.resource_type == other.resource_type and self.resource_id == other.resource_id
        )

    @staticmethod
    def resource_from_dict(resource_dict):
        assert isinstance(resource_dict, dict), 'resource_dict must be a dict'

        resource = PanoptesResource(resource_plugin=resource_dict['resource_plugin'],
                                    resource_site=resource_dict['resource_site'],
                                    resource_class=resource_dict['resource_class'],
                                    resource_subclass=resource_dict['resource_subclass'],
                                    resource_type=resource_dict['resource_type'],
                                    resource_id=resource_dict['resource_id'],
                                    resource_endpoint=resource_dict['resource_endpoint'],
                                    resource_ttl=int(resource_dict['resource_metadata'].get(
                                        '_resource_ttl', const.RESOURCE_MANAGER_RESOURCE_EXPIRE)),
                                    resource_creation_timestamp=resource_dict['resource_creation_timestamp'], )

        for metadata_key in resource_dict['resource_metadata']:
            resource.add_metadata(metadata_key, resource_dict['resource_metadata'][metadata_key])

        return resource


class PanoptesResourceSet(object):
    """
    An (un-ordered) set of PanoptesResources
    """

    def __init__(self):
        self.__data = dict()
        self.__data['resources'] = set()
        self.__data['resource_set_creation_timestamp'] = time()
        self.__data['resource_set_schema_version'] = '0.1'

    def add(self, resource):
        """
        Add a resource to the set

        Args:
            resource (PanoptesResource): The resource to add

        Returns:
            None

        """
        assert resource and isinstance(resource, PanoptesResource), 'resource must an instance of PanoptesResource'
        self.__data['resources'].add(resource)

    def remove(self, resource):
        """
        Remove a resource from the set

        Args:
            resource (PanoptesResource): The resource to remove

        Returns:
            None

        """
        assert resource and isinstance(resource, PanoptesResource), 'resource must an instance of PanoptesResource'
        self.__data['resources'].remove(resource)

    def get_resources_by_site(self):
        resources_by_site = dict()
        for resource in self.resources:
            if not resources_by_site.get(resource.resource_site, None):
                resources_by_site[resource.resource_site] = list()

            resources_by_site[resource.resource_site].append(resource)

        return resources_by_site

    @property
    def resources(self):
        """
        Return the list of resources in this set

        Returns:
            list: The set of resources

        """
        return self.__data['resources']

    @property
    def resource_set_creation_timestamp(self):
        return self.__data['resource_set_creation_timestamp']

    @resource_set_creation_timestamp.setter
    def resource_set_creation_timestamp(self, timestamp):
        """
            Sets the timestamp of the

        Args:
            timestamp (int, float): The new timestamp

        Returns:
            None

        Raises:
            AssertionError: If the passed timestamp is not an int or float or is too old or too much in the future, \
            an AssertionError would be raised

        """
        assert PanoptesValidators.valid_timestamp(timestamp), 'timestamp should be an Unix epoch int|float not more  ' \
                                                              'than 7 days old or more than 60 seconds in the future'
        self.__data['resource_set_creation_timestamp'] = timestamp

    @property
    def resource_set_schema_version(self):
        return self.__data['resource_set_schema_version']

    def __iter__(self):
        return iter(self.__data['resources'])

    def next(self):
        """
        Returns the next resource in the set

        Returns:
            PanoptesResource: The next resource in the set
        """
        return next(iter(self.__data['resources']))

    @property
    def json(self):
        """
        The JSON representation of the resource set

        Returns:
            str: The JSON representation of the resource set
        """
        return json.dumps(self.__data, cls=PanoptesResourceEncoder)

    def __repr__(self):
        return str(self.__data['resources'])

    def __len__(self):
        return len(self.__data['resources'])


class PanoptesResourcesKeyValueStore(PanoptesKeyValueStore):
    """
    A custom Key/Value store for Panoptes resources
    """
    redis_group = const.RESOURCE_MANAGER_REDIS_GROUP

    def __init__(self, panoptes_context):
        super(PanoptesResourcesKeyValueStore, self).__init__(panoptes_context,
                                                             const.RESOURCE_MANAGER_KEY_VALUE_NAMESPACE)


class PanoptesResourceStore(object):
    """
    This class implements methods to fetch resources from Redis and create in-memory objects of the same
    """
    _regex_key = re.compile(
            '(^plugin\|)(?P<plugin>.*?)(\|site\|)(?P<site>.*?)(\|class\|)(?P<class>.*?)(\|subclass\|)('
            '?P<subclass>.*?)(\|type\|)(?P<type>.*?)(\|id\|)(?P<id>.*?)(\|endpoint\|)(?P<endpoint>.*?$)')
    _regex_timestamp = re.compile('timestamp\|(?P<timestamp>(\d{10}\.\d{1,2}))\|?(?P<meta>.*)$')
    _regex_meta = re.compile('(meta\.)(.*?)(\|)(.*?)(\||$)')

    def __init__(self, panoptes_context):
        self.__panoptes_context = panoptes_context
        try:
            self.__kv = self.__panoptes_context.get_kv_store(PanoptesResourcesKeyValueStore)
        except Exception as e:
            raise e

    def get_resources(self, site=None, plugin_name=None):
        """
        Get all resources from the Redis store

        The serialization that this expects in Redis is:

        key: 'resource:resource_site:<sitename>:resource_class:<class>:resource_subclass:<subclass>:resouce_type
        :<type>:resource_id:<id>:resource_endpoint:endpoint'
        value: 'timestamp:<unix epoch>{:meta:<key name>:<value>}*'

        Args:
            site (str, None): An optional string which filters resources by site
            plugin_name (str, None): An optional string which filters resources by the plugin they were discovered
                                     through

        Returns:
            PanoptesResourceSet: All resources fetched from the Redis store
        """
        assert site is None or PanoptesValidators.valid_nonempty_string(site), 'site should be None or a non-empty str'
        assert plugin_name is None or PanoptesValidators.valid_nonempty_string(plugin_name), \
            'plugin_signature should be None or a non-empty str'

        logger = self.__panoptes_context.logger
        resources = PanoptesResourceSet()

        key_namespace = ''

        if plugin_name:
            key_namespace += 'plugin|' + plugin_name + '|'

        if site:
            key_namespace += 'site|' + site + '|'

        key_namespace += '*'

        logger.info('Trying to get all resources under key namespace "%s"' % key_namespace)

        start = time()

        for key in self.__kv.find_keys(pattern=key_namespace):
            logger.debug('Attempting to get resource under key "%s"' % key)

            try:
                resource = self.get_resource(key)
                logger.debug('Found resource "%s"' % resource)
            except:
                logger.exception('Error trying to get "%s", skipping resource' % key)
                continue

            resources.add(resource)

        end = time()
        logger.info('Fetched %d resource(s) in %.2f seconds' % (len(resources), end - start))

        return resources

    def get_resource(self, resource_key):
        assert PanoptesValidators.valid_nonempty_string(resource_key), 'resource_key must be a non-empty str or unicode'

        logger = self.__panoptes_context.logger

        try:
            value = self.__kv.get(resource_key)
        except Exception as e:
            logger.exception('Error trying to get value from key-value store for key "%s"' % resource_key)
            raise PanoptesResourceError('Error trying to fetch value for key "%s": %s ' % (resource_key, repr(e)))

        if not value:
            logger.exception('Error -- No resource found for key "%s"' % resource_key)
            raise PanoptesResourceError('No resource found for key "%s"' % resource_key)

        return self._deserialize_resource(resource_key, value)

    def add_resource(self, plugin_signature, resource):
        assert PanoptesValidators.valid_nonempty_string(plugin_signature), 'plugin_signature must be a non-empty str'
        assert resource and isinstance(resource, PanoptesResource), 'resource must be a non-empty instance of ' \
                                                                    'PanoptesResource'
        key, value = self._serialize_resource(resource)

        try:
            self.__kv.set(key, value, expire=int(resource.resource_ttl))
        except Exception as e:
            raise PanoptesResourceError('Error trying to add resource "%s": %s' % (resource, str(e)))

    def delete_resource(self, plugin_signature, resource):
        assert plugin_signature and isinstance(plugin_signature, str), 'plugin_signature must be a non-empty str'
        assert resource and isinstance(resource, PanoptesResource), 'resource must be a non-empty instance of ' \
                                                                    'PanoptesResource'

        key, value = self._serialize_resource(resource)

        try:
            self.__kv.delete(key)
        except Exception as e:
            raise PanoptesResourceError('Error trying to delete resource "%s": %s' % (resource, str(e)))

    @staticmethod
    def _serialize_resource(resource):
        assert resource and isinstance(resource, PanoptesResource), 'resource must be a non-empty instance of ' \
                                                                    'PanoptesResource'

        key = resource.serialization_key

        value = 'timestamp|' + str(resource.resource_creation_timestamp) + '|'

        for metadata_key in resource.resource_metadata:
            value += 'meta.' + metadata_key + '|' + resource.resource_metadata[metadata_key] + '|'

        return key, value

    def _deserialize_resource(self, key, value):
        """
        This method parses the key/value pair to create a PanoptesResource

        Args:
            key (str): The serialized key
            value (str): The serialized value

        Returns:
            PanoptesResource: A PanoptesResource object which represented by the serialized key and value
        """
        parsed_key = self.__class__._regex_key.search(key)

        if not parsed_key:
            raise PanoptesResourceError('Resource key "%s" does not match pattern' % key)

        parsed_value = self.__class__._regex_timestamp.search(value)

        if not parsed_value:
            raise PanoptesResourceError('Value for resource "%s" does not match pattern' % value)

        meta = self.__class__._regex_meta.findall(parsed_value.group('meta'))
        resource_metadata = dict()
        for m in meta:
            resource_metadata[m[1]] = m[3]

        resource = PanoptesResource(resource_site=parsed_key.group('site'),
                                    resource_class=parsed_key.group('class'),
                                    resource_subclass=parsed_key.group('subclass'),
                                    resource_type=parsed_key.group('type'),
                                    resource_id=parsed_key.group('id'),
                                    resource_endpoint=parsed_key.group('endpoint'),
                                    resource_plugin=parsed_key.group('plugin'),
                                    resource_ttl=int(resource_metadata.get('_resource_ttl',
                                                                           const.RESOURCE_MANAGER_RESOURCE_EXPIRE)),
                                    resource_creation_timestamp=float(parsed_value.group('timestamp')))

        for k in resource_metadata:
            resource.add_metadata(k, resource_metadata[k])

        return resource


class PanoptesResourceDSL(object):
    """
    The class implements a SQL WHERE clause like resource selection/filtering DSL

    The following operators are supported: =, !=, eq, ne, LIKE, AND, OR, NOT, IN

    Please see the 'Examples' and 'Notes' for usage and constraints

    Args:
        query(str): The query to select/filter resources
        panoptes_context(PanoptesContext): The PanoptesContext associated with the calling process

    Returns:
        PanoptesResourceDSL

    Examples:
        resource_site = "dc1"
        resource_site = "dc1" AND resource_class = "network"
        resource_metadata.os_version LIKE "4%" OR resource_site NOT IN ("dc1", "dc2")

    Notes:
        * All the values in the resource selection are treated as strings and should be quoted with double quotes
        * NOT operator can be used in front of LIKE and IN:
            * resource_site NOT IN ("dc1", "dc2")
            * resource_site NOT LIKE 'dc%'
        * Grouping by parenthesis 'e.g. NOT (resource_site = "dc1" OR resource_site = "dc2")' is currently not supported
    """

    def __init__(self, query, panoptes_context):
        assert query and isinstance(query, string_types), 'query must be an instance of str or unicode'
        assert panoptes_context and isinstance(panoptes_context,
                                               PanoptesContext), 'panoptes_context must be an instance of str or ' \
                                                                 'unicode'
        self._query = query
        logger = panoptes_context.logger

        logger.info('Parsing query expression: %s' % query)
        try:
            tokens = self._parse_query()
        except ParseException as e:
            logger.error('Error in parsing expression "%s": %s' % (query, str(e)))
            raise e
        else:
            logger.debug('Tokens = %s' % tokens)
            self._tokens = tokens

    def _parse_query(self):
        """
        Defines and parses the Resource DSL based on the query associated with the PanoptesResourceDSL object

        Returns:
            list: The list of tokens parsed

        Raises:
            ParseException: This exception is raised if any parsing error occurs
        """
        resource_fields = oneOf(
                'resource_site resource_class resource_subclass resource_type resource_id resource_endpoint',
                caseless=True)
        resource_metadata = CaselessLiteral('resource_metadata') + Literal('.') + Word(alphanums + '_')

        and_ = CaselessKeyword('AND').setParseAction(upcaseTokens)
        or_ = CaselessKeyword('OR').setParseAction(upcaseTokens)
        not_ = CaselessKeyword('NOT').setParseAction(upcaseTokens)
        in_ = CaselessKeyword('IN').setParseAction(upcaseTokens)
        like_ = CaselessKeyword('LIKE').setParseAction(upcaseTokens)

        operators = oneOf("= != eq ne", caseless=True).setParseAction(upcaseTokens)

        query_expression = Forward()

        query_l_val = (resource_fields | resource_metadata).setParseAction(downcaseTokens)
        query_r_val = QuotedString(quoteChar='"', escChar='\\')

        query_condition = Group(
                (query_l_val + operators + query_r_val) |
                (query_l_val + Optional(not_) + like_ + query_r_val) |
                (query_l_val + Optional(not_) + in_ + '(' + delimitedList(query_r_val) + ')')
        )

        query_expression << query_condition - ZeroOrMore((and_ | or_) - query_condition)

        try:
            tokens = query_expression.parseString(self._query, parseAll=True)
        except ParseException as e:
            raise e

        return tokens

    @property
    def tokens(self):
        """
        Return the tokens parsed from the query

        Returns:
            pyparsing.Parser: The tokens parsed from the query

        """
        return self._tokens

    @property
    def sql(self):
        """
        The property returns a part of the WHERE clause based on the input query which is compatible with SQLite

        Returns:
            str: A partial WHERE clause

        """
        metadata_sql = ''
        intersect_sql_clause = ''
        union_sql_clause = ''
        sql = ''
        metadata_first_clause = True

        for i in range(0, len(self._tokens)):
            token = self._tokens[i]
            if isinstance(token, ParseResults):
                if token[0] == 'resource_metadata':
                    metadata_sql_clause = '('
                    metadata_sql_clause += ('resource_metadata.key = ' + '"' + token[2] + '"' +
                                            ' AND resource_metadata.value ')
                    if token[3] == 'NOT':
                        metadata_sql_clause += token[3] + ' ' + token[4] + ' ' + self._process_rval(token[5:])
                    else:
                        metadata_sql_clause += token[3] + ' ' + self._process_rval(token[4:])
                    metadata_sql_clause += ')'
                    if metadata_first_clause:
                        metadata_first_clause = False
                        metadata_sql += metadata_sql_clause
                    else:
                        if i > 0:
                            partial_sql_clause = (
                                'SELECT resource_metadata.id FROM resources,resource_metadata WHERE ' +
                                metadata_sql_clause + ' AND resource_metadata.id = resources.id ')
                            if self._tokens[i - 1] == 'AND':
                                intersect_sql_clause += 'INTERSECT ' + partial_sql_clause
                            elif self._tokens[i - 1] == 'OR':
                                union_sql_clause += 'UNION ' + partial_sql_clause
                else:
                    sql += 'resources.' + token[0] + ' '
                    if token[1] == 'NOT':
                        sql += token[1] + ' ' + token[2] + ' ' + self._process_rval(token[3:])
                    else:
                        sql += token[1] + ' ' + self._process_rval(token[2:])
            else:
                if isinstance(self._tokens[i + 1], ParseResults) and self._tokens[i + 1][0] != 'resource_metadata':
                    sql += ' ' + token + ' '

        if metadata_sql:
            sql += ' AND (' + metadata_sql + ')'
            where_clause = ("(SELECT resource_metadata.id FROM resources,resource_metadata " +
                            "WHERE (" + sql + " AND resource_metadata.id = resources.id) " +
                            union_sql_clause +
                            intersect_sql_clause +
                            "GROUP BY resource_metadata.id " +
                            "ORDER BY resource_metadata.id" +
                            ") AS filtered_resources")

            final_sql = ('SELECT resources.*,group_concat(key,"|"),group_concat(value,"|") ' +
                         'FROM ' + where_clause + ', resources, resource_metadata ' +
                         'WHERE resources.id = filtered_resources.id ' +
                         'AND resource_metadata.id = filtered_resources.id ' +
                         'GROUP BY resource_metadata.id')
        else:
            final_sql = ('SELECT resources.*, group_concat(key,"|"), group_concat(value,"|") ' +
                         'FROM resources ' +
                         'LEFT JOIN resource_metadata ON resources.id = resource_metadata.id ' +
                         "WHERE (" + sql + ") " +
                         union_sql_clause +
                         intersect_sql_clause +
                         "GROUP BY resource_metadata.id " +
                         "ORDER BY resource_metadata.id")
        return final_sql

    @staticmethod
    def _process_rval(rval):
        """
        This method creates SQL for the right hand side of an expression

        Args:
            rval (list):

        Returns:
            str: The partial SQL statement
        """
        sql = ''
        if rval[0] == '(':
            sql += '(' + ','.join(['"' + x + '"' for x in rval[1:len(rval) - 1]]) + ')'
        else:
            sql = '"' + rval[0] + '"'

        return sql


class PanoptesResourceCache:
    """
    This class implements an in-memory cache of PanoptesResources which can be queried using the DSL

    Internally, it fetches all Panoptes Resources from Redis and creates an in-memory SQLite DB which can be queried

    Args:
        panoptes_context (PanoptesContext): The Panoptes Context associated with the calling Plugin Scheduler
    """

    def __init__(self, panoptes_context):
        self._resources = None
        self._db = None
        self._cursor = None
        self._panoptes_context = panoptes_context
        self._resources_store = None
        self._cached_resources = dict()
        self._lock = threading.Lock()

    def get_resources(self, query):
        """
        Returns a resource set filtered on the provided Resource DSL query

        Args:
            query (str): The Panoptes Resource DSL filter to apply

        Returns:
            PanoptesResourceSet: The resource set containing the filtered resources
        """
        logger = self._panoptes_context.logger

        logger.info('Going to query with resource filter "%s"' % query)
        resource_filter = PanoptesResourceDSL(query, self._panoptes_context)
        logger.debug('SQL for resource filter "%s": %s' % (query, resource_filter.sql))

        if resource_filter.sql in self._cached_resources:
            logger.debug('Returning cached resources')
            return self._cached_resources[resource_filter.sql]

        try:
            self._cursor.execute(resource_filter.sql)
        except Exception as e:
            raise PanoptesResourceCacheException(
                    'Error trying to execute resource filter SQL "%s": %s' % (resource_filter.sql, str(e)))

        resource_set = PanoptesResourceSet()
        rows = self._cursor.fetchall()

        for row in rows:
            logger.debug('For resource filter "%s", found resource %s' % (query, row))
            resource = PanoptesResource(resource_site=row[1], resource_class=row[2], resource_subclass=row[3],
                                        resource_type=row[4], resource_id=row[5], resource_endpoint=row[6],
                                        resource_plugin=row[7])

            try:
                # Columns 8 & 9 contain the metadata keys and values respectively
                if row[8] and row[9]:
                    metadata_keys = row[8].split('|')
                    metadata_values = row[9].split('|')
                    for i in range(len(metadata_keys)):
                        resource.add_metadata(metadata_keys[i], metadata_values[i])
            except Exception as e:
                logger.error(
                    'Either resource metadata key or value are not correct, skipping resource "%s": %s' % (
                        row, str(e)))
                continue

            resource_set.add(resource)

        with self._lock:
            self._cached_resources[resource_filter.sql] = resource_set
        logger.info('For resource filter "%s", found %d resources' % (query, len(resource_set)))
        logger.debug('For resource filter "%s", returning resource set: %s' % (query, resource_set))
        return resource_set

    def setup_resource_cache(self):
        """
        Gets resources from a PanoptesResourceStore (Redis) and populates the in-memory SQLite DB

        Returns:
            None

        Raises:
            Exception: Passes through any exceptions that happen during fetching or population of resources
        """
        logger = self._panoptes_context.logger

        try:
            self._resources_store = PanoptesResourceStore(self._panoptes_context)
        except Exception as e:
            logger.error('Error while setting up PanoptesResourceStore: %s' % repr(e))
            raise PanoptesResourceCacheException("Error while setting up resources_store")

        try:
            self._create_db()
        except Exception as e:
            logger.error('Error while setting up the in-memory SQLite DB: %s' % repr(e))
            raise PanoptesResourceCacheException("Error while setting up the in-memory SQLite DB")

        logger.info('Attempting to get all resources')
        try:
            self._resources = self._resources_store.get_resources()
        except Exception as e:
            logger.error('Error while getting resources from PanoptesResourceStore: %s' % repr(e))
            raise PanoptesResourceCacheException("Error while getting resources from PanoptesResourceStore")

        logger.info('Got %d resources' % len(self._resources))
        logger.debug('Resources: %s' % self._resources)

        for resource in self._resources:
            self._cursor.execute(
                    '''
                    INSERT INTO resources(resource_site,
                                          resource_class,
                                          resource_subclass,
                                          resource_type,
                                          resource_id,
                                          resource_endpoint,
                                          resource_plugin)
                    VALUES (?,?,?,?,?,?,?)
                    ''''',
                    (resource.resource_site,
                     resource.resource_class,
                     resource.resource_subclass,
                     resource.resource_type,
                     resource.resource_id,
                     resource.resource_endpoint,
                     resource.resource_plugin)
            )

            rowid = self._cursor.lastrowid
            logger.debug('Inserted row into database with rowid %d' % rowid)
            logger.debug('Going to update metadata associated with the resource')

            for key in resource.resource_metadata:
                self._cursor.execute(
                        '''
                        INSERT INTO resource_metadata(id, key, value) VALUES(?,?,?)
                        ''',
                        (rowid, key, resource.resource_metadata[key])
                )
        self._db.commit()
        # Invalidate cached resources
        with self._lock:
            self._cached_resources = dict()

        logger.info("Created DB")

    def _create_db(self):
        """
        Creates the in-memory SQLite DB and stores reference to the cursor

        Returns:
            None
        """
        logger = self._panoptes_context.logger

        logger.info('Creating in-memory SQLite DB')
        self._db = sqlite3.connect(':memory:')
        self._cursor = self._db.cursor()
        self._cursor.arraysize = const.RESOURCE_CACHE_DB_CURSOR_SIZE

        logger.info('Creating tables in SQLite DB')
        self._cursor.execute(
                '''
                CREATE TABLE resources
                (id INTEGER PRIMARY KEY AUTOINCREMENT,
                resource_site VARCHAR(255) NOT NULL,
                resource_class VARCHAR(255) NOT NULL,
                resource_subclass VARCHAR(255) NOT NULL,
                resource_type VARCHAR(255) NOT NULL,
                resource_id VARCHAR(255) NOT NULL,
                resource_endpoint VARCHAR(255) NOT NULL,
                resource_plugin VARCHAR(255) NOT NULL
                )
                '''
        )

        self._cursor.execute(
                '''
                CREATE TABLE resource_metadata
                (id INTEGER,
                key VARCHAR(255) NOT NULL,
                value VARCHAR(255) NOT NULL,
                FOREIGN KEY(id) REFERENCES resources(id)
                )
                '''
        )

        self._cursor.execute('CREATE INDEX index_resource_site on resources (resource_site);')
        self._cursor.execute('CREATE INDEX index_resource_class on resources (resource_class);')
        self._cursor.execute('CREATE INDEX index_resource_subclass on resources (resource_subclass);')
        self._cursor.execute('CREATE INDEX index_resource_type on resources (resource_type);')
        self._cursor.execute('CREATE INDEX index_resource_id on resources (resource_id);')
        self._cursor.execute('CREATE INDEX index_resource_endpoint on resources (resource_endpoint);')
        self._cursor.execute('CREATE INDEX index_resource_plugin on resources (resource_plugin);')
        self._cursor.execute('CREATE INDEX index_metadata_id on resource_metadata (id);')
        self._cursor.execute('CREATE INDEX index_metadata_key on resource_metadata (key);')
        self._cursor.execute('CREATE INDEX index_metadata_value on resource_metadata (value);')

        self._db.commit()

        logger.info('Finished creating tables in SQLite DB')

    def close_resource_cache(self):
        """
        Closes the connection to the in-memory SQLite DB

        Returns:
            None
        """
        if self._db:
            self._db.close()
        else:
            self._panoptes_context.logger.error("Attempted to close connection to SQLite DB that was not open")


class PanoptesResourceEncoder(json.JSONEncoder):
    # https://github.com/PyCQA/pylint/issues/414
    def default(self, o):  # pylint: disable=E0202
        if isinstance(o, set):
            return list(o)
        if isinstance(o, PanoptesResource):
            return o.__dict__['_PanoptesResource__data']
        return json.JSONEncoder.default(self, o)
